try:
    from . import _build_meta

    __version__ = _build_meta.__version__
    __banner__ = f'wwwpy v{__version__}'
except:
    __version__ = 'unknown'
    __banner__ = 'wwwpy version unknown'

__all__ = ['__version__']



from typing import List

from wwwpy.common.designer.element_library import AttributeDef, Help, EventDef, ElementDef

ad_style = AttributeDef('style', Help('Inline style.',
                                      'https://developer.mozilla.org/en-US/docs/Web/API/HTMLElement/style'))
ad_class = AttributeDef('class', Help('A space-separated list of the classes of the element.',
                                      'https://developer.mozilla.org/en-US/docs/Web/API/Element/classList'))

ad_value = AttributeDef('value', Help('The value of the element.', ''))
ad_disabled = AttributeDef('disabled', Help('Whether the control is disabled.', ''), boolean=True)
ad_readonly = AttributeDef('readonly', Help('Whether the control is read-only.', ''), boolean=True)
ad_autofocus = AttributeDef('autofocus',
                            Help('Whether the control should have input focus when the page loads.', ''),
                            boolean=True)
ad_required = AttributeDef('required', Help('Whether the control is required for form submission.', ''),
                           boolean=True)
ad_placeholder = AttributeDef('placeholder', Help('A hint to the user of what can be entered in the field.', ''))
ad_name = AttributeDef('name', Help('Name of the control, useful for form submission.', ''))
ed_click = EventDef('click', Help('The element was clicked.',
                                  'https://developer.mozilla.org/en-US/docs/Web/API/Element/click_event'))
ed_dblclick = EventDef('dblclick', Help('The element was double-clicked.',
                                        'https://developer.mozilla.org/en-US/docs/Web/API/Element/dblclick_event'))
ed_input = EventDef('input', Help('The input event fires when the value of the element has been changed '
                                  'as a direct result of a user action',
                                  'https://developer.mozilla.org/en-US/docs/Web/API/Element/input_event'))
ed_keydown = EventDef('keydown', Help('The keydown event is fired when a key is pressed.'
                                      , 'https://developer.mozilla.org/en-US/docs/Web/API/Element/keydown_event'))
ed_keyup = EventDef('keyup', Help('The keyup event is fired when a key is released.',
                                  'https://developer.mozilla.org/en-US/docs/Web/API/Element/keyup_event'))
ed_change = EventDef('change', Help(
    'The change event is fired when a change to the element\'s value is committed by the user.',
    'https://developer.mozilla.org/en-US/docs/Web/API/Element/change_event'))

ed_pointerdown = EventDef('pointerdown', Help('The pointerdown event is fired when a pointer becomes active.',
                                              'https://developer.mozilla.org/en-US/docs/Web/API/Element/pointerdown_event'))

ed_pointerup = EventDef('pointerup', Help('The pointerup event is fired when a pointer is no longer active.',
                                          'https://developer.mozilla.org/en-US/docs/Web/API/Element/pointerup_event'))
ed_pointermove = EventDef('pointermove', Help('The pointermove event is fired when a pointer changes coordinates.',
                                              'https://developer.mozilla.org/en-US/docs/Web/API/Element/pointermove_event'))


def _insert_common_to_all(elements: List[ElementDef]):
    attrs = list(reversed([ad_style, ad_class]))
    events = list(reversed([ed_keydown, ed_keyup, ed_pointerdown, ed_pointerup, ed_pointermove, ]))

    for element in elements:
        ea = element.attributes
        for attr in attrs:
            if ea.get(attr.name) is None:
                ea.insert(0, attr)

        ee = element.events
        for event in events:
            if ee.get(event.name) is None:
                ee.append(event)



import asyncio
import logging
from pathlib import Path

logging.getLogger().setLevel(logging.DEBUG)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

logging.info('Loading module')


def open_browser_once():
    open_browser_once.__code__ = (lambda: None).__code__
    import webbrowser
    webbrowser.open('http://localhost:8000')


# scope # https://github.com/django/asgiref/blob/main/asgiref/typing.py#L66
async def app(scope, receive, send):
    func = _scopes.get(scope['type'], None)
    if func:
        await func(scope, receive, send)
    else:
        logger.info(f"scope type not handled: {scope['type']}")


async def scope_lifespan(scope, receive, send):
    open_browser_once()
    logger.info(f"lifespan scope: {scope}")
    await send({'type': 'lifespan.startup.complete'})

    message = await receive()
    if message['type'] == 'lifespan.shutdown':
        await send({'type': 'lifespan.shutdown.complete'})


async def scope_http(scope, receive, send):
    path = scope['path']
    headers = dict(scope['headers'])
    logger.info(f"http path: {path}")
    if path == '/':
        await serve_file('index.html', 'text/html', send)
    elif path == '/favicon.ico':  # png actually
        await serve_file('favicon.png', 'image/png', send)
    elif path == '/post':
        body = await _all_body(receive)
        await send_response(200, body, 'text/plain', send)
    else:
        await send_response(200, b'Hello, world! ' + f'you requested `{path}`'.encode('utf-8'),
                            'text/plain', send)


async def scope_websocket(scope, receive, send):
    logger.info(f"websocket scope: {scope}")
    scope_path = scope['path']
    headers = dict(scope['headers'])
    if scope_path != '/echo':
        return
    await send({'type': 'websocket.accept'})
    send_text = lambda t: send({'type': 'websocket.send', 'text': t})

    async def send_hello():
        try:
            await send_text('producing hello messages every 2 seconds')
            hello_count = 0
            while True:
                hello_count += 1
                await asyncio.gather(send_text(f'hello {hello_count}'), asyncio.sleep(2))
        except asyncio.CancelledError:
            pass

    hello_task = asyncio.create_task(send_hello())

    try:
        while True:
            message = await receive()
            if message['type'] == 'websocket.receive':
                await send_text(f"echo -> {message.get('text')}")
            elif message['type'] == 'websocket.disconnect':
                break
    finally:
        # Cancel the background task when the connection is closed
        hello_task.cancel()


async def serve_file(filename, content_type, send):
    try:
        content = (Path(__file__).parent / filename).read_bytes()
        await send_response(200, content, content_type, send)
    except FileNotFoundError:
        await send_response(404, b'File not found', 'text/plain', send)


async def send_response(status, body, content_type, send):
    await send({
        'type': 'http.response.start',
        'status': status,
        'headers': [[b'content-type', content_type.encode()], ],
    })
    await send({'type': 'http.response.body', 'body': body, })


_scopes = {
    'http': scope_http,
    'websocket': scope_websocket,
    'lifespan': scope_lifespan,
}


async def _all_body(receive):
    body = b""
    more_body = True

    while more_body:
        message = await receive()
        body += message.get("body", b"")
        more_body = message.get("more_body", False)
    return body



from __future__ import annotations

import asyncio
import logging
import traceback
from inspect import iscoroutinefunction

import js
from js import console

import wwwpy.common.reloader as reloader
from wwwpy.common import _no_remote_infrastructure_found_text, files, _remote_module_not_found_html
from wwwpy.common.asynclib import create_task_safe
from wwwpy.remote.designer import dev_mode as dm
from wwwpy.remote.websocket import setup_websocket

logger = logging.getLogger(__name__)


async def entry_point(dev_mode: bool = False):
    # from wwwpy.common.tree import print_tree
    # print_tree('/wwwpy_bundle')
    import wwwpy
    console.log(wwwpy.__banner__)
    await setup_websocket()
    dm.set_active(dev_mode)
    if dev_mode:
        await dm.activate()

    await _invoke_browser_main()


def _reload():
    async def reload():
        console.log('reloading')
        # for p in Path(files._bundle_path).iterdir():
        #     if p.name == 'wwwpy':
        #         continue
        #     reloader.unload_path(str(p))
        reloader.unload_path(files._bundle_path, skip_wwwpy=True)
        await _invoke_browser_main()

    asyncio.create_task(reload())


async def _invoke_browser_main():
    try:
        console.log('invoke_browser_main')

        try:
            js.document.documentElement.innerHTML = ''
            js.document.head.innerHTML = ''
            for attr in js.document.documentElement.attributes:
                js.document.documentElement.removeAttributeNode(attr)

            js.document.body.innerText = f'Importing the "remote" package...'
            import remote
            if hasattr(remote, 'main'):
                if iscoroutinefunction(remote.main):
                    await remote.main()
                else:
                    remote.main()
        except Exception as e:
            explain_text = _no_remote_infrastructure_found_text
            if isinstance(e, ModuleNotFoundError) and e.name == 'remote':
                explain_text = _remote_module_not_found_html
            js.console.error(f'Exception: {e}')
            _show_exception(e, explain_text)
            from wwwpy.server.designer import rpc
            create_task_safe(rpc.on_exception_string(traceback.format_exc()))
    finally:
        await create_task_safe(_show_dev_mode())


async def _show_dev_mode():
    if dm.is_active():
        from wwwpy.remote.designer.ui import dev_mode_component
        dev_mode_component.show()  # todo it looks like it can take ~ 1s or more; investigate,
        # maybe micropip installing rope has to do with it


def _reset_document():
    pass
    # the following code is very controversial.
    # js.document.open()
    # js.document.write('<!DOCTYPE html><html></html>')
    # js.document.close()


def _show_exception(e, html: str):
    js.document.body.innerHTML = html
    js.document.body.insertAdjacentHTML('beforeend', '<br><br>')

    exc_trace = traceback.format_exc()
    exc_str = 'Exception: ' + str(e) + '\n\n' + exc_trace + '\n\n'
    pre = js.document.createElement('pre')
    pre.innerText = exc_str
    js.document.body.appendChild(pre)



import logging


def set_log_level(log_level: dict[str, str]):
    tr = translate_names(log_level)
    for module, level in tr.items():
        logging.getLogger(module).setLevel(level)


def translate_names(log_level):
    translated = {}
    for module, level in log_level.items():
        name = logging.getLevelName(level)
        if isinstance(name, int):
            translated[module] = name
    return translated



import importlib

required = {
    ('pytest',),
    ('xvirt', 'pytest-xvirt'),
    ('playwright',),
    ('pytest_playwright', 'pytest-playwright')
}
missing_pip = set()
installed = set()
for tup in required:

    pip_name = tup[0] if len(tup) == 1 else tup[1]
    try:
        importlib.import_module(tup[0])
        installed.add(pip_name)
    except ImportError:
        missing_pip.add(pip_name)

if len(missing_pip) > 0:
    msg = ('You need to install the following packages to use this plugin:\n   '
           + ', '.join(missing_pip)
           + '\n\nYou can install them by running:\n   `pip install ' + ' '.join(missing_pip) +
           '`\n\n'
           + 'Packages already installed: ' + ', '.join(installed)
           )
    raise ImportError(msg)



from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass, field
from functools import partial
from typing import Optional, Tuple, List

import js
from js import document, console, Event, HTMLElement, window
from pyodide.ffi import create_proxy

import wwwpy.remote.component as wpc
from wwwpy.common import state, modlib
from wwwpy.common.designer import element_library
from wwwpy.common.designer.code_edit import add_component, AddResult, AddFailed
from wwwpy.common.designer.element_library import Help, ElementDef
from wwwpy.common.designer.element_path import ElementPath
from wwwpy.common.designer.html_edit import Position
from wwwpy.common.designer.html_locator import path_to_index
from wwwpy.remote import dict_to_js
from wwwpy.remote.designer import element_path
from wwwpy.remote.designer.drop_zone import DropZone, DropZoneHover
from wwwpy.remote.designer.global_interceptor import GlobalInterceptor, InterceptorEvent
from wwwpy.remote.designer.helpers import _element_lbl, _help_button, info_link, _help_url
from wwwpy.remote.designer.ui.property_editor import PropertyEditor
from wwwpy.remote.designer.ui.window_component import WindowComponent
from wwwpy.remote.designer.ui.window_component import new_window, Geometry
from wwwpy.server.designer import rpc
from . import filesystem_tree
from .help_icon import HelpIcon  # noqa
from .mailto_component import MailtoComponent

logger = logging.getLogger(__name__)


@dataclass
class ToolboxState:
    geometry: Tuple[int, int, int, int] = field(default=(30, 200, 400, 330))
    toolbox_search: str = ''
    selected_element_path: Optional[ElementPath] = None


@dataclass
class MenuMeta:
    label: str
    html: str
    always_visible: bool = False
    p_element: js.HTMLElement = None
    help: Help = None


def menu(label, always_visible=False):
    def wrapped(fn, label=label):
        help = None
        if isinstance(label, Help):
            help = label
            label = label.description
        fn.label = label
        fn.meta = MenuMeta(label, label, always_visible, help=help)
        return fn

    return wrapped


class ToolboxComponent(wpc.Component, tag_name='wwwpy-toolbox'):
    _title: js.HTMLDivElement = wpc.element()
    body: HTMLElement = wpc.element()
    inputSearch: js.HTMLInputElement = wpc.element()
    _window: WindowComponent = wpc.element()
    property_editor: PropertyEditor = wpc.element()
    _select_element_btn: js.HTMLElement = wpc.element()
    _select_clear_btn: js.HTMLElement = wpc.element()
    components_marker = '-components-'

    @property
    def visible(self) -> bool:
        return self._window.element.style.display != 'none'

    @visible.setter
    def visible(self, value: bool):
        self._window.element.style.display = 'block' if value else 'none'

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
<style>
.two-column-layout {
  display: flex;
  flex-wrap: wrap;
}

.two-column-layout p {
  width: calc(50% - 10px); /* 50% width minus half of the gap */
  margin: 0 20px 10px 0; /* Right margin creates gap between columns */
}

.two-column-layout p:nth-child(even) {
  margin-right: 0; /* Removes right margin for every even child */
}
</style>     
<wwwpy-window data-name='_window'>
    <div slot='title' data-name="_title" style='text-align: center'></div>
    <div  style="text-align: center; padding: 8px">     
        <button data-name="_select_element_btn">Select element...</button>   
        <button data-name="_select_clear_btn">Clear selection</button>   
        <wwwpy-property-editor data-name="property_editor"></wwwpy-property-editor>        
        <p><input data-name='inputSearch' type='search' placeholder='type to filter...'></p>
        <div data-name='body' class='two-column-layout'></div>
    </div>   
</wwwpy-window>         
"""
        import wwwpy
        self._title.innerText = f'toolbox - {wwwpy.__banner__}'

    def connectedCallback(self):
        self._manage_toolbox_state()
        self._window.closable = False
        attrs = [v for k, v in vars(self.__class__).items() if hasattr(v, 'label')]
        self._all_items: List[MenuMeta] = []

        def add_p(menu_meta: MenuMeta, callback):  # callback can be async
            self._all_items.append(menu_meta)
            p: js.HTMLElement = document.createElement('p')
            p.style.color = 'white'
            menu_meta.p_element = p
            p.innerHTML = menu_meta.html
            p.addEventListener('click', create_proxy(callback))

        def add_comp(element_def: ElementDef):

            def _on_hover(drop_zone: DropZone | None):
                console.log(f'pointed dropzone: {drop_zone}')
                msg = (f'Insert new {element_def.tag_name}')
                if drop_zone:
                    pos = 'before' if drop_zone.position == Position.beforebegin else 'after'
                    # msg += f' at {drop_zone.position.name} of {drop_zone.element.tagName}'
                    msg += f' {pos} {drop_zone.element.tagName}'
                else:
                    msg += ' ... select a dropzone on the page.'
                self.property_editor.message1div.innerHTML = msg

            async def _start_drop_for_comp(event):
                _on_hover(None)
                res = await _drop_zone_start_selection_async(_on_hover, whole=False)
                logger.debug(f'_start_drop_for_comp res={res}')
                if res:
                    self.property_editor.set_state_selection_active()
                    await self._process_dropzone(res, element_def)
                else:
                    await self._canceled()

            element_html = f'<span style="cursor: pointer">{element_def.tag_name}</span> {_help_button(element_def)}'
            add_p(MenuMeta(element_def.tag_name, element_html), _start_drop_for_comp)

        for member in attrs:
            if member.meta.label == self.components_marker:
                [add_comp(ele_def) for ele_def in element_library.element_library().elements]
            else:
                help = member.meta.help
                if help:
                    help_html = '' if not help.url else info_link(help.url)
                    element_html = f'{help.description} {help_html}'
                    add_p(MenuMeta(help.description, element_html), partial(member, self))
                else:
                    add_p(member.meta, partial(member, self))
        self._update_toolbox_elements()

    async def _process_dropzone(self, drop_zone: DropZone, element_def: ElementDef):
        el_path = element_path.element_path(drop_zone.element)

        if not el_path:
            window.alert(f'No component found for dropzone!')
            return

        logger.debug(f'element_path={el_path}')
        file = modlib._find_module_path(el_path.class_module)
        old_source = file.read_text()

        path_index = path_to_index(el_path.path)
        add_result = add_component(old_source, el_path.class_name, element_def, path_index, drop_zone.position)

        if isinstance(add_result, AddResult):
            logger.debug(f'write_module_file len={len(add_result.source_code)} el_path={el_path}')
            new_element_path = ElementPath(el_path.class_module, el_path.class_name, add_result.node_path,
                                           el_path.origin)
            self._toolbox_state.selected_element_path = new_element_path
            write_res = await rpc.write_module_file(el_path.class_module, add_result.source_code)
            logger.debug(f'write_module_file res={write_res}')
        elif isinstance(add_result, AddFailed):
            js.alert('Sorry, an error occurred while adding the component.')
            _open_error_reporter_window(
                'Error report data:\n\n' + add_result.exception_report_b64,
                title='Error report add_component - wwwpy'
            )
            # pre1: js.HTMLElement = js.document.createElement('pre')
            # pre1.innerText = self.path.read_text()
            # win.element.append(pre1)

    def _manage_toolbox_state(self):
        self._toolbox_state = state._restore(ToolboxState)
        self.inputSearch.value = self._toolbox_state.toolbox_search
        self._window.set_geometry(Geometry(*self._toolbox_state.geometry))

        def on_toolbar_geometry_change():
            g = self._window.geometry()
            acceptable_geometry = g.width > 100 and g.height > 100 and g.top > 0 and g.left > 0
            if acceptable_geometry:
                self._toolbox_state.geometry = self._window.geometry()

        self._window.geometry_change_listeners.append(on_toolbar_geometry_change)
        self._restore_selected_element_path()

    def inputSearch__input(self, e: Event):
        self._toolbox_state.toolbox_search = self.inputSearch.value
        self._update_toolbox_elements()

    def _update_toolbox_elements(self):
        def search_match(meta: MenuMeta):
            search = self.inputSearch.value.lower()
            lbl = meta.label.lower()
            starts_with = search.startswith(' ') and lbl.startswith(search.lstrip())
            ends_with = search.endswith(' ') and lbl.endswith(search.rstrip())
            if starts_with or ends_with:
                return True
            return search in lbl

        self.body.innerHTML = ''
        for meta in self._all_items:
            p = meta.p_element
            if meta.always_visible or search_match(meta):
                self.body.appendChild(p)

    async def _select_clear_btn__click(self, e: Event):
        self._toolbox_state.selected_element_path = None
        self._restore_selected_element_path()

    async def _select_element_btn__click(self, e: Event):
        no_comp = 'Select an element on the page to be inspected and edited...'
        self.property_editor.message1div.innerHTML = no_comp

        def _on_hover(drop_zone: DropZone | None):
            msg = no_comp
            if drop_zone:
                msg = 'Click to select ' + _element_lbl(drop_zone.element)
            self.property_editor.message1div.innerHTML = msg
            console.log(f'pointed dropzone: {drop_zone}')

        res = await _drop_zone_start_selection_async(_on_hover, whole=True)
        if res:
            self.property_editor.set_state_selection_active()
            self._toolbox_state.selected_element_path = element_path.element_path(res.element)
        else:
            await self._canceled()
        self._restore_selected_element_path()

    # @menu(Help("Open errore reporter", ''))
    # async def _open_error_reporter(self, e: Event):
    #     _open_error_reporter_window('test error report body')

    @menu(Help('Create new component', _help_url('add_component')))
    async def _add_new_component(self, e: Event):
        if js.window.confirm('Add new component file?\nIt will be added to your "remote" folder.'):
            res = await rpc.add_new_component()
            js.window.alert(res)

    @menu(Help("Explore local filesystem", _help_url('remote_filesystem')))
    async def _browse_local_filesystem(self, e: Event):
        from wwwpy.remote.designer.ui.dev_mode_component import DevModeComponent
        filesystem_tree.show_explorer(DevModeComponent.instance.root_element())

    @menu(Help("Python console", ''))
    async def _python_console(self, e: Event):
        from wwwpy.remote.designer.ui.dev_mode_component import DevModeComponent
        w1 = new_window('Python console')
        from wwwpy.remote.designer.ui.python_console import PythonConsoleComponent
        w1.element.append(PythonConsoleComponent().element)
        DevModeComponent.instance.root_element().append(w1.element)

    @menu(components_marker)
    def _drop_zone_start(self, e: Event):
        assert False, 'Just a placeholder'

    # @menu('handle global click')
    def _handle_global_click(self, e: Event):
        # add input
        def global_click(event: Event):
            event.preventDefault()
            event.stopImmediatePropagation()
            event.stopPropagation()
            console.log('global click', event.element)
            if self._global_click:
                document.removeEventListener('click', self._global_click, True)
                self._global_click = None

        self._global_click = create_proxy(global_click)
        document.addEventListener('click', self._global_click, True)

    def _restore_selected_element_path(self):
        element_path = self._toolbox_state.selected_element_path
        if element_path:
            console.log(f'restoring selected element path: {element_path}')
            if not element_path.valid():
                element_path = None

        if element_path:
            self._select_clear_btn.style.visibility = 'visible'
        else:
            self._select_clear_btn.style.visibility = 'hidden'

        self.property_editor.selected_element_path = element_path

    async def _canceled(self):
        self.property_editor.message1div.innerHTML = 'Operation canceled'
        await asyncio.sleep(2)
        self.property_editor.message1div.innerHTML = ''


def is_inside_toolbar(element: HTMLElement | None):
    if not element:
        return False

    orig = element
    # loop until the root element and see if it is the toolbar
    res = False
    while element:
        if element.tagName.lower() == ToolboxComponent.component_metadata.tag_name:
            res = True
            break
        element = element.parentElement

    return res


def _default_drop_zone_accept(drop_zone: DropZone):
    element = drop_zone.element
    name = element.tagName.lower()
    from wwwpy.remote.designer.ui.dev_mode_component import DevModeComponent
    root_elements = name == 'body' or name == 'html'
    wwwpy_elements = name == DevModeComponent.component_metadata.tag_name or is_inside_toolbar(element)
    accept = not (root_elements or wwwpy_elements)
    return accept


async def _drop_zone_start_selection_async(on_hover: DropZoneHover, whole: bool) -> Optional[DropZone]:
    from wwwpy.remote.designer.drop_zone import drop_zone_selector

    event = asyncio.Event()
    result = []

    def intercept_ended(ev: InterceptorEvent):
        console.log('intercept_ended', ev.target)
        ev.preventAndStop()
        selected = drop_zone_selector.stop()
        ev.uninstall()
        if selected:
            console.log(
                f'\nselection accepted position {selected.position.name}'
                f'\ntarget: ', selected.element,
                '\nparent: ', selected.element.parentElement,
                '\nevent: ', ev.event,
                '\ncomposedPath: ', ev.event.composedPath(),
            )
            result.append(selected)
        else:
            console.log('intercept_ended - canceled')
        event.set()

    GlobalInterceptor(intercept_ended, 'pointerdown').install()

    click_inter = GlobalInterceptor(lambda ev: ev.preventAndStop(), 'click')
    click_inter.install()
    drop_zone_selector.start_selector(on_hover, _default_drop_zone_accept, whole=whole)
    await event.wait()
    console.log('drop_zone_selector event ended')

    async def _click_inter_uninstall():
        await asyncio.sleep(0.5)
        click_inter.uninstall()

    asyncio.ensure_future(_click_inter_uninstall())

    if len(result) == 0:
        return None

    return result[0]


def _open_error_reporter_window(body: str, title='Error reporter - wwwpy'):
    win = new_window(title, closable=True)
    mailto_component = MailtoComponent()
    mailto_component.subject = title
    mailto_component.recipient = 'simone.giacomelli@gmail.com'
    mailto_component.body = body
    mailto_component.text_content = 'Click here to send an email with the error report'
    mailto_component.element.style.display = 'block'
    mailto_component.element.style.textAlign = 'center'
    div: js.HTMLElement = js.document.createElement('div')

    action = '<h3>Please, help us improve wwwpy by sending this error report</h3>'
    div.insertAdjacentHTML('beforeend', action)
    div.append(mailto_component.element)
    div.style.margin = '15px'
    win.element.append(div)
    div.insertAdjacentHTML('beforeend', '<br>')
    js.document.body.append(win.element)

    from wwwpy.remote.designer.ui.dev_mode_component import DevModeComponent
    DevModeComponent.instance.root_element().append(win.element)



from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, TypeVar, Union

_S = TypeVar('_S')
"""Success type variable"""
_F = TypeVar('_F')
"""Failure type variable"""


@dataclass(frozen=True)
class Result(Generic[_S, _F]):
    _value: Union[_S, _F]
    _is_success: bool

    @staticmethod
    def success(value: _S) -> Success[_S, _F]:
        return Success(value)

    @staticmethod
    def failure(error: _F) -> Failure[_S, _F]:
        return Failure(error)

    @property
    def is_success(self) -> bool:
        return self._is_success

    @property
    def is_failure(self) -> bool:
        return not self._is_success

    def get_or_null(self) -> _S | None:
        return self._value if self.is_success else None

    def error_or_null(self) -> _F | None:
        return self._value if self.is_failure else None

    def get_or_throw(self) -> _S:
        if self.is_success:
            return self._value

        exc = self._value
        if not isinstance(self._value, BaseException):
            exc = Exception(self._value)
        raise exc

    def __repr__(self) -> str:
        return f"Success({self._value})" if self.is_success else f"Failure({self._value})"


class Success(Result[_S, _F]):
    def __init__(self, value: _S):
        super().__init__(value, True)


class Failure(Result[_S, _F]):
    def __init__(self, error: _F):
        super().__init__(error, False)


def main():
    from typing import get_origin, get_args

    result_type = Result[int, str]
    print("Origin:", get_origin(result_type))  # Should print: <class '__main__.Result'>
    print("Args:", get_args(result_type))  # Should print: (<class 'int'>, <class 'str'>)

    success = Result.success(42)
    print("Instance type args:", get_args(success.__orig_class__))  # Extract generics from an instance


if __name__ == '__main__':
    main()



from __future__ import annotations

import logging

import js
from js import Array, Element, document

from wwwpy.common.designer.element_path import ElementPath, Origin
from wwwpy.common.designer.html_locator import Node
from wwwpy.remote.component import get_component

logger = logging.getLogger(__name__)


def element_path(element: Element) -> ElementPath | None:
    """Returns an instance of  """

    path = []
    while element:
        if is_instance_of(element, js.ShadowRoot):
            element = element.host
        component = get_component(element)
        if component:
            clazz = component.__class__
            return ElementPath(clazz.__module__, clazz.__name__, path, Origin.live)
        if element == document.body:
            return None

        parent = element.parentNode
        child_index = Array.prototype.indexOf.call(parent.children, element) if parent else -1
        attributes = {attr.name: attr.value for attr in element.attributes}
        path.insert(0, Node(element.tagName.lower(), child_index, attributes))
        element = parent

    return None


_instanceof = js.eval('(i,t) => i instanceof t')


def is_instance_of(instance, js_type):
    return _instanceof(instance, js_type)



import base64
import builtins
import dataclasses
import enum
import importlib
import json
import re
import sys
import types
import typing
from dataclasses import is_dataclass
from datetime import datetime
from typing import Any, Type, get_origin, get_args, TypeVar, List, Optional

from wwwpy.common import result

T = TypeVar('T')

class SerializationError(Exception):
    """Custom exception for serialization errors with path tracking."""
    def __init__(self, message: str, path: List[str]):
        self.path = path
        path_str = ".".join(str(p) for p in path) if path else "root"
        super().__init__(f"At {path_str}: {message}")


def serialize(obj: T, cls: Type[T], path: Optional[List[str]] = None) -> Any:
    """
    Serialize an object based on its type annotation.

    Args:
        obj: The object to serialize
        cls: The expected type of the object
        path: The current path in the object tree for error reporting

    Returns:
        The serialized representation of the object

    Raises:
        SerializationError: If serialization fails with detailed path information
    """
    path = path or [f"{cls}"]

    try:
        optional_type = _get_optional_type(cls)
        if optional_type:
            if obj is None:
                return None
            return serialize(obj, optional_type, path)

        origin = typing.get_origin(cls)

        # Handle union types
        if _is_union_type(origin):
            return _serialize_union(obj, cls, path)

        # Handle Result types
        if origin is result.Result:
            return _serialize_result(obj, cls, path)

        # Type checking
        if origin is not None:
            if not isinstance(obj, origin):
                raise SerializationError(
                    f"Expected object of type {origin}, got {type(obj).__name__}", path
                )
        elif not isinstance(obj, cls):
            raise SerializationError(
                f"Expected object of type {cls.__name__}, got {type(obj).__name__}", path
            )

        # Handle different types
        if is_dataclass(obj):
            return _serialize_dataclass(obj, cls, path)
        elif isinstance(obj, list):
            return _serialize_list(obj, cls, path)
        elif isinstance(obj, tuple):
            return _serialize_tuple(obj, cls, path)
        elif isinstance(obj, dict):
            return _serialize_dict(obj, cls, path)
        elif isinstance(obj, datetime):
            return obj.isoformat()
        elif isinstance(obj, bytes):
            return base64.b64encode(obj).decode('utf-8')
        elif isinstance(obj, (int, float, str, bool)):
            return obj
        elif isinstance(obj, enum.Enum):
            return serialize(obj.value, type(obj.value), path)
        elif obj is None:
            return None
        else:
            raise SerializationError(f"Unsupported type: {type(obj).__name__}", path)

    except SerializationError:
        # Just re-raise SerializationError as it already has path information
        raise
    except Exception as e:
        # Wrap other exceptions with context
        raise SerializationError(
            f"Serialization failed: {str(e)}", path
        ) from e


def _serialize_union(obj: Any, cls: Type, path: List[str]) -> Any:
    """Handle serialization of Union types."""
    args = set(get_args(cls))
    obj_type = type(obj)

    if obj_type not in args:
        valid_types = ", ".join(t.__name__ for t in args)
        raise SerializationError(
            f"Expected object of type {valid_types}, got {obj_type.__name__}", path
        )

    try:
        obj_ser = serialize(obj, obj_type, path)
        return [str(obj_type), obj_ser]
    except Exception as e:
        raise SerializationError(
            f"Failed to serialize union type {obj_type.__name__}", path
        ) from e


def _serialize_result(obj: Any, cls: Type, path: List[str]) -> Any:
    """Handle serialization of Result types."""
    args = set(get_args(cls))

    try:
        _assert_valid_result_type(type(obj))
    except ValueError as e:
        raise SerializationError(str(e), path) from e

    obj_type = type(obj._value)
    if obj_type not in args:
        valid_types = ", ".join(t.__name__ for t in args)
        raise SerializationError(
            f"Expected object of type Result {valid_types}, got {obj_type.__name__}", path
        )

    value_path = path + ["value"]
    try:
        obj_ser = serialize(obj._value, obj_type, value_path)
        return [str(type(obj)), obj_ser]
    except Exception as e:
        raise SerializationError(
            f"Failed to serialize Result value of type {obj_type.__name__}", path
        ) from e


def _serialize_dataclass(obj: Any, cls: Type, path: List[str]) -> dict:
    """Handle serialization of dataclass objects."""
    result = {}
    field_types = typing.get_type_hints(cls)

    for field_name, field_type in field_types.items():
        field_path = path + [field_name]
        try:
            field_value = getattr(obj, field_name)
            result[field_name] = serialize(field_value, field_type, field_path)
        except Exception as e:
            if not isinstance(e, SerializationError):
                raise SerializationError(
                    f"Failed to serialize field '{field_name}' of type {field_type.__name__}",
                    field_path
                ) from e
            raise

    return result


def _serialize_list(obj: list, cls: Type, path: List[str]) -> list:
    """Handle serialization of list objects."""
    item_type = typing.get_args(cls)[0]
    result = []

    for i, item in enumerate(obj):
        item_path = path + [str(i)]
        try:
            result.append(serialize(item, item_type, item_path))
        except Exception as e:
            if not isinstance(e, SerializationError):
                raise SerializationError(
                    f"Failed to serialize list item at index {i}", item_path
                ) from e
            raise

    return result


def _serialize_tuple(obj: tuple, cls: Type, path: List[str]) -> list:
    """Handle serialization of tuple objects."""
    result = []
    item_types = get_args(cls)

    for i, (item, item_type) in enumerate(zip(obj, item_types)):
        item_path = path + [str(i)]
        try:
            result.append(serialize(item, item_type, item_path))
        except Exception as e:
            if not isinstance(e, SerializationError):
                raise SerializationError(
                    f"Failed to serialize tuple item at index {i}", item_path
                ) from e
            raise

    return result


def _serialize_dict(obj: dict, cls: Type, path: List[str]) -> dict:
    """Handle serialization of dictionary objects."""
    key_type, value_type = typing.get_args(cls)
    result = {}

    for key, value in obj.items():
        key_path = path + ["key"]
        value_path = path + [str(key)]

        try:
            serialized_key = serialize(key, key_type, key_path)
            serialized_value = serialize(value, value_type, value_path)
            result[serialized_key] = serialized_value
        except Exception as e:
            if not isinstance(e, SerializationError):
                raise SerializationError(
                    f"Failed to serialize dictionary entry with key '{key}'",
                    value_path
                ) from e
            raise

    return result


def _get_optional_type(cls):
    """Extract the type from Optional[Type]."""
    origin = typing.get_origin(cls)
    args = typing.get_args(cls)

    if origin is typing.Union:
        if type(None) in args and len(args) == 2:
            return next(arg for arg in args if arg is not type(None))
    return None


class DeserializationError(Exception):
    """Custom exception for deserialization errors with path tracking."""
    def __init__(self, message: str, path: List[str]):
        self.path = path
        path_str = ".".join(str(p) for p in path) if path else "root"
        super().__init__(f"At {path_str}: {message}")


def deserialize(data: Any, cls: Type[T], path: Optional[List[str]] = None) -> T:
    """
    Deserialize data into an object of the specified type.

    Args:
        data: The serialized data
        cls: The target type to deserialize into
        path: The current path in the object tree for error reporting

    Returns:
        The deserialized object

    Raises:
        DeserializationError: If deserialization fails with detailed path information
    """
    path = path or [f"{cls}"]

    try:
        optional_type = _get_optional_type(cls)
        if optional_type:
            if data is None:
                return None
            return deserialize(data, optional_type, path)

        origin = get_origin(cls)

        # Handle union types
        if _is_union_type(origin):
            return _deserialize_union(data, cls, path)

        # Handle Result types
        if origin is result.Result:
            return _deserialize_result(data, cls, path)

        # Handle different types
        if is_dataclass(cls):
            return _deserialize_dataclass(data, cls, path)
        elif origin == list or cls == list:
            return _deserialize_list(data, cls, path)
        elif origin == tuple or cls == tuple:
            return _deserialize_tuple(data, cls, path)
        elif origin == dict or cls == dict:
            return _deserialize_dict(data, cls, path)
        elif isinstance(data, list):
            return _deserialize_subclass_list(data, cls, origin, path)
        elif cls == datetime:
            try:
                return datetime.fromisoformat(data)
            except ValueError as e:
                raise DeserializationError(
                    f"Invalid datetime format: {data}", path
                ) from e
        elif cls == bytes:
            try:
                return base64.b64decode(data.encode('utf-8'))
            except Exception as e:
                raise DeserializationError(
                    f"Invalid base64 data", path
                ) from e
        elif cls in (int, float, str, bool):
            try:
                return cls(data)
            except (ValueError, TypeError) as e:
                raise DeserializationError(
                    f"Cannot convert {data} to {cls.__name__}", path
                ) from e
        elif issubclass(cls, enum.Enum):
            try:
                first_member = next(iter(cls))
                value = deserialize(data, type(first_member.value), path)
                return cls(value)
            except (ValueError, KeyError) as e:
                valid_values = [m.value for m in cls]
                raise DeserializationError(
                    f"Invalid enum value for {cls.__name__}. Expected one of {valid_values}", path
                ) from e
        elif cls is type(None):
            if data is not None:
                raise DeserializationError(
                    f"Expected None, got {type(data).__name__}", path
                )
            return None
        else:
            raise DeserializationError(f"Unsupported type: {cls.__name__}", path)

    except DeserializationError:
        # Just re-raise DeserializationError as it already has path information
        raise
    except Exception as e:
        # Wrap other exceptions with context
        raise DeserializationError(
            f"Deserialization failed: {str(e)}", path
        ) from e


def _deserialize_union(data: list, cls: Type, path: List[str]) -> Any:
    """Handle deserialization of Union types."""
    if not isinstance(data, list) or len(data) != 2:
        raise DeserializationError(
            f"Invalid union data format: expected [type_str, value], got {data}", path
        )

    args = set(get_args(cls))
    try:
        obj_type = _get_type_from_string(data[0])
    except ValueError as e:
        raise DeserializationError(
            f"Invalid type string: {data[0]}", path
        ) from e

    if obj_type not in args:
        valid_types = ", ".join(t.__name__ for t in args)
        raise DeserializationError(
            f"Type {obj_type.__name__} not in union {valid_types}", path
        )

    try:
        return deserialize(data[1], obj_type, path)
    except Exception as e:
        if not isinstance(e, DeserializationError):
            raise DeserializationError(
                f"Failed to deserialize union value of type {obj_type.__name__}", path
            ) from e
        raise


def _deserialize_result(data: list, cls: Type, path: List[str]) -> Any:
    """Handle deserialization of Result types."""
    if not isinstance(data, list) or len(data) != 2:
        raise DeserializationError(
            f"Invalid Result data format: expected [type_str, value], got {data}", path
        )

    args = list(get_args(cls))
    try:
        obj_type = _get_type_from_string(data[0])
        _assert_valid_result_type(obj_type)
    except ValueError as e:
        raise DeserializationError(str(e), path) from e

    value_path = path + ["value"]
    if obj_type is result.Success:
        try:
            des = deserialize(data[1], args[0], value_path)
            return result.Success(des)
        except Exception as e:
            if not isinstance(e, DeserializationError):
                raise DeserializationError(
                    f"Failed to deserialize Success value", value_path
                ) from e
            raise
    elif obj_type is result.Failure:
        try:
            des = deserialize(data[1], args[1], value_path)
            return result.Failure(des)
        except Exception as e:
            if not isinstance(e, DeserializationError):
                raise DeserializationError(
                    f"Failed to deserialize Failure value", value_path
                ) from e
            raise
    else:
        raise DeserializationError(
            f"Expected Result type (Success or Failure), got {obj_type.__name__}", path
        )


def _deserialize_dataclass(data: dict, cls: Type, path: List[str]) -> Any:
    """Handle deserialization of dataclass objects."""
    if not isinstance(data, dict):
        raise DeserializationError(
            f"Expected dict for dataclass, got {type(data).__name__}", path
        )

    args = {}
    field_types = typing.get_type_hints(cls)

    # Get dataclass fields to check for default values
    dc_fields = {field.name: field for field in dataclasses.fields(cls)}

    # Check for missing required fields (those without default values)
    missing_fields = [
        field for field in field_types
        if field not in data and field in dc_fields and
           not dc_fields[field].default is dataclasses.MISSING and
           not dc_fields[field].default_factory is dataclasses.MISSING
    ]

    if missing_fields:
        raise DeserializationError(
            f"Missing required fields: {', '.join(missing_fields)}", path
        )

    # Process available fields
    for field_name, field_type in field_types.items():
        if field_name in data:
            field_path = path + [field_name]
            args[field_name] = deserialize(data[field_name], field_type, field_path)

    # Create the dataclass instance
    try:
        return cls(**args)
    except TypeError as e:
        raise DeserializationError(f"Failed to create dataclass: {str(e)}", path) from e

def _deserialize_list(data: list, cls: Type, path: List[str]) -> list:
    """Handle deserialization of list objects."""
    if not isinstance(data, list):
        raise DeserializationError(
            f"Expected list, got {type(data).__name__}", path
        )

    item_type = get_args(cls)[0]
    result = []

    for i, item in enumerate(data):
        item_path = path + [str(i)]
        try:
            result.append(deserialize(item, item_type, item_path))
        except Exception as e:
            if not isinstance(e, DeserializationError):
                raise DeserializationError(
                    f"Failed to deserialize list item at index {i}", item_path
                ) from e
            raise

    return result


def _deserialize_tuple(data: list, cls: Type, path: List[str]) -> tuple:
    """Handle deserialization of tuple objects."""
    if not isinstance(data, list):
        raise DeserializationError(
            f"Expected list for tuple, got {type(data).__name__}", path
        )

    item_types = get_args(cls)
    if len(data) != len(item_types):
        raise DeserializationError(
            f"Expected tuple of length {len(item_types)}, got {len(data)}", path
        )

    result = []
    for i, (item, item_type) in enumerate(zip(data, item_types)):
        item_path = path + [str(i)]
        try:
            result.append(deserialize(item, item_type, item_path))
        except Exception as e:
            if not isinstance(e, DeserializationError):
                raise DeserializationError(
                    f"Failed to deserialize tuple item at index {i}", item_path
                ) from e
            raise

    return tuple(result)


def _deserialize_dict(data: dict, cls: Type, path: List[str]) -> dict:
    """Handle deserialization of dictionary objects."""
    if not isinstance(data, dict):
        raise DeserializationError(
            f"Expected dict, got {type(data).__name__}", path
        )

    key_type, value_type = get_args(cls)
    result = {}

    for key, value in data.items():
        key_path = path + ["key"]

        try:
            deserialized_key = deserialize(key, key_type, key_path)
            value_path = path + [str(key)]
            deserialized_value = deserialize(value, value_type, value_path)
            result[deserialized_key] = deserialized_value
        except Exception as e:
            if not isinstance(e, DeserializationError):
                raise DeserializationError(
                    f"Failed to deserialize dictionary entry with key '{key}'",
                    path + [str(key)]
                ) from e
            raise

    return result


def _deserialize_subclass_list(data: list, cls: Type, origin, path: List[str]) -> Any:
    """Handle deserialization of list subclasses."""
    # Check if cls is a subclass of list or has origin that is a subclass of list
    ok = (origin is not None and issubclass(origin, list)) or (origin is None and issubclass(cls, list))
    if ok:
        # For subclasses of list
        item_type = get_args(cls)[0] if get_args(cls) else Any
        items = [deserialize(item, item_type, path + [str(i)]) for i, item in enumerate(data)]
        # If origin is None, use cls directly
        list_class = origin if origin is not None else cls
        return list_class(items)
    else:
        # If cls is not a subclass of list, raise an error
        raise DeserializationError(
            f"Expected a subclass of list, got {cls.__name__}", path
        )

def _get_type_from_string(type_str):
    """Convert a string representation of a type to the actual type object."""
    try:
        # Use regex to extract the full type path
        match = re.search(r"<class '(.+)'>|(.+)", type_str)
        if not match:
            raise ValueError(f"Invalid type string format: {type_str}")

        full_path = match.group(1) or match.group(2)

        if full_path == 'NoneType':
            return type(None)

        # Check if it's a builtin type
        if hasattr(builtins, full_path):
            return getattr(builtins, full_path)

        # Split the path into parts
        parts = full_path.split('.')

        # The class name is the last part
        class_name = parts.pop()

        # The module name is everything else
        module_name = '.'.join(parts)

        # Import the module dynamically
        module = importlib.import_module(module_name)

        # Get the class from the module
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        raise ValueError(f"Failed to resolve type '{type_str}': {str(e)}")


def _assert_valid_result_type(typ):
    """Ensure the type is a valid Result type (Success or Failure)."""
    success = typ is result.Success
    failure = typ is result.Failure
    ok = success or failure
    if not ok:
        raise ValueError(f"Expected object of type Result (Success or Failure), got {typ.__name__}")


def _is_union_type_3_10(origin):
    """Check if a type origin is a Union type in Python 3.10+."""
    return origin is typing.Union or origin is types.UnionType


def _is_union_type_3_9(origin):
    """Check if a type origin is a Union type in Python 3.9."""
    return origin is typing.Union


# Select the appropriate Union check based on Python version
_is_union_type = _is_union_type_3_10 if sys.version_info >= (3, 10) else _is_union_type_3_9


def to_json(obj: Any, cls: Type[T]) -> str:
    """
    Serialize an object to a JSON string.

    Args:
        obj: The object to serialize
        cls: The expected type of the object

    Returns:
        A JSON string representation of the object
    """
    try:
        return json.dumps(serialize(obj, cls))
    except SerializationError as e:
        # Re-raise with a more descriptive message
        raise SerializationError(
            f"JSON serialization failed: {str(e)}", e.path
        ) from e


def from_json(json_str: str, cls: Type[T]) -> T:
    """
    Deserialize a JSON string into an object of the specified type.

    Args:
        json_str: The JSON string to deserialize
        cls: The target type to deserialize into

    Returns:
        The deserialized object
    """
    try:
        data = json.loads(json_str)
    except json.JSONDecodeError as e:
        raise DeserializationError(
            f"Invalid JSON format: {str(e)}", []
        ) from e

    try:
        return deserialize(data, cls)
    except DeserializationError as e:
        # Re-raise with a more descriptive message
        raise DeserializationError(
            f"JSON deserialization failed: {str(e)}", e.path
        ) from e


from typing import List

from wwwpy.common.designer.el_common import ad_value, ad_disabled, ad_readonly, ad_autofocus, ad_required, \
    ad_placeholder, ad_name, ed_click, ed_dblclick, ed_input, ed_change, _insert_common_to_all
from wwwpy.common.designer.element_library import ElementDef, Help, AttributeDef, NamedListMap


def _standard_elements_def() -> List[ElementDef]:
    res = [
        ElementDef(
            'button', 'js.HTMLButtonElement',
            help=Help('A clickable button.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/button'),
            attributes=[
                AttributeDef('type', Help('The type of the button.', ''),
                             values=['submit', 'reset', 'button'], default_value='button'),
                ad_value,
                ad_disabled,
                ad_autofocus,
            ],
            events=[ed_click, ed_dblclick],
        ),
        ElementDef(
            'input', 'js.HTMLInputElement',
            help=Help('A field for entering text.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input'),
            attributes=[
                AttributeDef('type', Help('The type of control to display. The default type is text.', ''),
                             values=['text', 'password', 'checkbox', 'radio', 'button', 'submit', 'reset', 'file',
                                     'hidden', 'image', 'date', 'datetime-local', 'month', 'time', 'week',
                                     'number', 'range', 'email', 'url', 'search', 'tel', 'color'],
                             default_value='text'),
                ad_value,
                ad_placeholder,
                ad_disabled,
                ad_readonly,
                ad_required,
                ad_name,
                AttributeDef('min', Help('The minimum value of the input field.', '')),
                AttributeDef('max', Help('The maximum value of the input field.', '')),
                AttributeDef('step', Help('The legal number intervals for the input field.', '')),
                AttributeDef('pattern', Help('A regular expression that the input\'s value is checked against.', '')),
                AttributeDef('autocomplete', Help(
                    'Lets web developers specify what if any permission the user agent has to provide automated assistance in filling out form field values, as well as guidance to the browser as to the type of information expected in the field.',
                    ''),
                             values=['on', 'off']),
                AttributeDef('autocapitalize', Help(
                    'Controls whether inputted text is automatically capitalized and, if so, in what manner.', ''),
                             values=['on', 'off']),
                ad_autofocus,
                AttributeDef('checked', Help('Whether the control is checked.', ''), boolean=True),
                AttributeDef('multiple', Help('Whether the user is allowed to enter more than one value.', ''),
                             boolean=True),

            ],
            events=[ed_input, ed_click, ed_change]
        ),
        ElementDef(
            'textarea', 'js.HTMLTextAreaElement',
            help=Help(
                'A multi-line text input control.',
                'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/textarea'
            ),
            attributes=[
                AttributeDef(
                    'cols',
                    Help('The visible width of the text control, in average character widths.', ''),
                    default_value='20'
                ),
                AttributeDef(
                    'rows',
                    Help('The number of visible text lines for the control.', ''),
                    default_value='2'
                ),
                ad_disabled,
                ad_readonly,
                ad_required,
                ad_placeholder,
                ad_autofocus,
                AttributeDef(
                    'maxlength',
                    Help('The maximum number of characters that the user can enter.', '')
                ),
                AttributeDef(
                    'minlength',
                    Help('The minimum number of characters that the user can enter.', '')
                ),
                AttributeDef(
                    'wrap',
                    Help('How the control wraps text.', ''),
                    values=['soft', 'hard', 'off'],
                    default_value='soft'
                ),
                AttributeDef(
                    'autocomplete',
                    Help('Whether the control has autocomplete enabled.', ''),
                    values=['on', 'off'],
                    default_value='on'
                ),
                AttributeDef(
                    'spellcheck',
                    Help('Whether spell checking is enabled for the control.', ''),
                    values=['true', 'false'],
                    default_value='true'
                ),
                AttributeDef(
                    'form',
                    Help('The form element that the textarea is associated with (its id).', '')
                ),
                ad_name,
            ],
            events=[ed_input, ed_change, ],
        ),
        ElementDef(
            'div', 'js.HTMLDivElement',
            help=Help('A generic container element.',
                      'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/div')
        ),
        ElementDef(
            'br', 'js.HTMLBRElement',
            help=Help('It produces a line break in text (carriage-return).',
                      'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/br')
        ),
        ElementDef(
            'progress', 'js.HTMLProgressElement',
            help=Help('A progress bar.',
                      'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/progress'),
            attributes=[
                AttributeDef('value', Help('The current value of the progress bar.', '')),
                AttributeDef('max', Help('The maximum value of the progress bar.', '')),
            ],
            events=[ed_click, ]
        ),
        ElementDef(
            'select', 'js.HTMLSelectElement',
            help=Help(
                'A control that provides a menu of options.',
                'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/select'
            ),
            attributes=[
                ad_name,
                ad_disabled,
                AttributeDef(
                    'multiple',
                    Help('Whether multiple options can be selected.', ''),
                    boolean=True
                ),
                ad_required,
                AttributeDef(
                    'size',
                    Help('Number of visible options.', ''),
                    default_value='1'
                ),
                ad_autofocus,
                AttributeDef(
                    'form',
                    Help('The form element that the select is associated with (its id).', '')
                ),
            ],
            events=[ed_change, ed_input],

        ),
        ElementDef(
            'meter', 'js.HTMLMeterElement',
            help=Help(
                'A scalar measurement within a known range (e.g., disk usage, battery level).',
                'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meter'
            ),
            attributes=NamedListMap([
                AttributeDef('value', Help('The current numeric value.', '')),
                AttributeDef('min', Help('The lower numeric bound of the measured range. Default is 0.', '')),
                AttributeDef('max', Help('The upper numeric bound of the measured range. Default is 1.', '')),
                AttributeDef('low', Help('The upper numeric bound of the low end of the measured range.', '')),
                AttributeDef('high', Help('The lower numeric bound of the high end of the measured range.', '')),
                AttributeDef('optimum', Help('The optimal numeric value.', '')),
                AttributeDef('form', Help('Associates the meter with a form element.', '')), ]
            ),
            events=NamedListMap([ed_click, ]),
        ),
        ElementDef('a', 'js.HTMLAnchorElement',
                   help=Help('A hyperlink.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/a'),
                   attributes=[AttributeDef('href', Help('The URL of the hyperlink.',
                                                         'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/a'))])
        ,
        ElementDef('img', 'js.HTMLImageElement',
                   help=Help('An image.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/img'), attributes=[
                AttributeDef('src', Help('The source URL of the image.',
                                         'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/img')),
                AttributeDef('alt', Help('Alternative text for the image.',
                                         'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/img'))])
        ,
        ElementDef('option', 'js.HTMLOptionElement', help=Help('An option in a select list.',
                                                               'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/option'),
                   attributes=[AttributeDef('value', Help('The value of the option.',
                                                          'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/option')),
                               AttributeDef('disabled', Help('Disables the option.',
                                                             'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/option'),
                                            boolean=True)])
        ,
        ElementDef('form', 'js.HTMLFormElement',
                   help=Help('A form.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/form'), attributes=[
                AttributeDef('action', Help('The URL to process the form submission.',
                                            'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/form')),
                AttributeDef('method', Help('The HTTP method for form submission.',
                                            'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/form'),
                             values=['get', 'post'], default_value='get')])
        ,
        ElementDef('script', 'js.HTMLScriptElement',
                   help=Help('A script.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/script'),
                   attributes=[AttributeDef('type', Help('The MIME type of the script.',
                                                         'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/script'))])
        ,
        ElementDef('pre', 'js.HTMLPreElement',
                   help=Help('Preformatted text.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/pre'),
                   events=[ed_click, ed_dblclick])
        ,
        ElementDef('code', 'js.HTMLElement',
                   help=Help('Computer code.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/code'),
                   events=[ed_click, ed_dblclick])

    ]

    _add_additional(res)

    for r in res:
        r.gen_html = _generateHtml

    _insert_common_to_all(res)
    return res


_element_additional = [
    # tag_name, python_type, description, url
    ('h1', 'js.HTMLHeadingElement', 'A top-level heading.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h1'),
    ('h2', 'js.HTMLHeadingElement', 'A level-2 heading.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h2'),
    ('h3', 'js.HTMLHeadingElement', 'A level-3 heading.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h3'),
    ('h4', 'js.HTMLHeadingElement', 'A level-4 heading.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h4'),
    ('h5', 'js.HTMLHeadingElement', 'A level-5 heading.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h5'),
    ('h6', 'js.HTMLHeadingElement', 'A level-6 heading.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h6'),
    ('p', 'js.HTMLParagraphElement', 'A paragraph.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/p'),
    ('span', 'js.HTMLSpanElement', 'A generic inline container.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/span'),
    ('strong', 'js.HTMLElement', 'A strong emphasis.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/strong'),
    ('em', 'js.HTMLElement', 'An emphasized text.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/em'),
    ('ul', 'js.HTMLUListElement', 'An unordered list.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/ul'),
    ('ol', 'js.HTMLOListElement', 'An ordered list.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/ol'),
    ('li', 'js.HTMLLIElement', 'A list item.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/li'),
    ('table', 'js.HTMLTableElement', 'A table.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/table'),
    ('thead', 'js.HTMLTableSectionElement', 'A table header.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/thead'),
    ('tbody', 'js.HTMLTableSectionElement', 'A table body.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/tbody'),
    ('tr', 'js.HTMLTableRowElement', 'A table row.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/tr'),
    ('th', 'js.HTMLTableCellElement', 'A table header cell.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/th'),
    ('td', 'js.HTMLTableCellElement', 'A table data cell.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/td'),
    ('label', 'js.HTMLLabelElement', 'A label.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/label'),
    ('optgroup', 'js.HTMLOptGroupElement', 'A group of options in a select list.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/optgroup'),
    ('hr', 'js.HTMLHRElement', 'A thematic break.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/hr'),
    ('nav', 'js.HTMLElement', 'A navigation section.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/nav'),
    ('header', 'js.HTMLElement', 'A header section.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/header'),
    ('footer', 'js.HTMLElement', 'A footer section.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/footer'),
    ('main', 'js.HTMLElement', 'A main section.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/main'),
    ('section', 'js.HTMLElement', 'A section.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/section'),
    ('article', 'js.HTMLElement', 'An article.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/article'),
    ('aside', 'js.HTMLElement', 'An aside section.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/aside'),
    ('figure', 'js.HTMLFigureElement', 'A figure.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/figure'),
    ('figcaption', 'js.HTMLElement', 'A figure caption.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/figcaption'),
    ('details', 'js.HTMLDetailsElement', 'A details section.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/details'),
    ('summary', 'js.HTMLElement', 'A summary.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/summary'),
    ('video', 'js.HTMLVideoElement', 'A video.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/video'),
    ('audio', 'js.HTMLAudioElement', 'An audio.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/audio'),
    ('source', 'js.HTMLSourceElement', 'A source.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/source'),
    ('canvas', 'js.HTMLCanvasElement', 'A canvas.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/canvas'),
    (
        'iframe', 'js.HTMLIFrameElement', 'An iframe.',
        'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/iframe'),
    ('embed', 'js.HTMLEmbedElement', 'An embed.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/embed'),
    (
        'object', 'js.HTMLObjectElement', 'An object.',
        'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/object'),
    ('map', 'js.HTMLMapElement', 'A map.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/map'),
    ('style', 'js.HTMLStyleElement', 'A style.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/style'),
    ('title', 'js.HTMLTitleElement', 'A title.', 'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/title'),
    ('template', 'js.HTMLTemplateElement', 'A template.',
     'https://developer.mozilla.org/en-US/docs/Web/HTML/Element/template'),
]


def _add_additional(elements: List[ElementDef]):
    already_present = {e.tag_name for e in elements}
    for tag_name, python_type, description, url in _element_additional:
        if tag_name not in already_present:
            elements.append(ElementDef(tag_name, python_type, help=Help(description, url)))


def _generateHtml(element_def: ElementDef, name: str) -> str:
    tag_name = element_def.tag_name

    def _def(placeHolder=False, add='', inner=None):
        def inner_func():
            pl = '' if not placeHolder else f' placeholder="{name}"'
            add1 = '' if not add else f' {add}'
            content = name if inner is None else inner
            return f'<{tag_name} data-name="{name}"{pl}{add1}>{content}</{tag_name}>'

        return inner_func

    func = {
        'button': _def(),
        'label': _def(),
        'div': _def(),
        'br': lambda: '<br>',
        'input': lambda: f'<input data-name="{name}" placeholder="{name}">',
        'img': lambda: f'<img data-name="{name}" src="https://images.unsplash.com/photo-1517331156700-3c241d2b4d83?ixlib=rb-1.2.1&ixid=eyJhcHBfaWQiOjEyMDd9&auto=format&fit=crop&w=400&q=90" alt="{name}">',
        'progress': lambda: f'<progress data-name="{name}" value="70" max="100">70%</progress>',
        'textarea': _def(placeHolder=True, inner='', add='rows="6" wrap="off" style="width: 100%"'),
        'select': _def(inner='''
            <option value="option1">Option 1</option>
            <option value="option2">Option 2</option>
            <option value="option3">Option 3</option>
        '''),
        'meter': _def(add='value="0.85" min="0" max="1" low="0.3" high="0.7" optimum="0.8"'),
        'span': _def(),
        'a': _def(),
    }
    gen_html = func.get(tag_name, None)
    html = '\n' + gen_html() if gen_html else '' + ElementDef.default_gen_html(element_def, name)
    return html

# [
#     _comp('input', 'js.HTMLInputElement', '<input type="text" data-name="#name#" value="#name#">'),
#     _comp('button', 'js.HTMLButtonElement', '<button data-name="#name#">#name#</button>'),
#     _comp('textarea', 'js.HTMLTextAreaElement', '<textarea data-name="#name#">#name#</textarea>'),
#     _comp('select', 'js.HTMLSelectElement',
#           '<select data-name="#name#"><option>#name#</option></select>'),
#     _comp('div', 'js.HTMLDivElement', '<div data-name="#name#">#name#</div>'),
#     _comp('p', 'js.HTMLParagraphElement', '<p data-name="#name#">#name#</p>'),
#     _comp('h1', 'js.HTMLHeadingElement', '<h1 data-name="#name#">#name#</h1>'),
#     _comp('h2', 'js.HTMLHeadingElement', '<h2 data-name="#name#">#name#</h2>'),
#     _comp('h3', 'js.HTMLHeadingElement', '<h3 data-name="#name#">#name#</h3>'),
#     _comp('a', 'js.HTMLAnchorElement', '<a href="#" data-name="#name#">#name#</a>'),
#     _comp('img', 'js.HTMLImageElement', '<img src="#" alt="#name#" data-name="#name#">'),
#     _comp('ul', 'js.HTMLUListElement', '<ul data-name="#name#"><li>#name#</li></ul>'),
#     _comp('ol', 'js.HTMLOListElement', '<ol data-name="#name#"><li>#name#</li></ol>'),
#     _comp('li', 'js.HTMLLIElement', '<li data-name="#name#">#name#</li>'),
#     _comp('table', 'js.HTMLTableElement', '<table data-name="#name#"><tr><td>#name#</td></tr></table>'),
#     _comp('form', 'js.HTMLFormElement', '<form data-name="#name#">#name#</form>'),
#     _comp('label', 'js.HTMLLabelElement', '<label data-name="#name#">#name#</label>'),
#
# ]



import importlib
from wwwpy.common.rpc.func_registry import FuncMeta


class Invoker:
    def __init__(self, module: FuncMeta):
        self.module = module
        self.module_type = _load_package(module.name)

    def __getitem__(self, item):
        return _Invocable(getattr(self.module_type, item))


class _Invocable:
    def __init__(self, func):
        self.func = func


def _load_package(package_name: str):
    try:
        return importlib.import_module(package_name)
    except ImportError:
        print(f"Package '{package_name}' could not be loaded.")
        return None



"""Module library for finding module paths and roots."""
from __future__ import annotations

import os
import sys
from pathlib import Path

import logging

logger = logging.getLogger(__name__)

def _find_module_path(module_name: str) -> Path | None:
    """Finds the path of a module without loading it."""
    parts = module_name.split('.')
    for sys_path in map(Path, sys.path):
        module_path = sys_path.joinpath(*parts)
        py_file = module_path.with_suffix('.py')
        init_file = module_path / '__init__.py'

        if py_file.is_file():
            return py_file
        if module_path.is_dir() and init_file.is_file():
            return init_file

    return None

def _find_package_directory(package_name) -> Path | None:
    """Finds the directory of a package without loading it."""
    path = _find_module_path(package_name)
    if not path:
        return None
    if path.name != '__init__.py':
        raise ValueError(f'package {package_name} is not a package')
    return path.parent


def _find_module_root(fqn, full_path):
    parts = fqn.split('.')
    relative_path = os.path.join(*parts[:-1]) + '.py'
    full_path = os.path.abspath(full_path)
    index = full_path.rfind(relative_path)
    if index == -1:
        return None
    return full_path[index:]



from pathlib import Path
import logging
import configparser

logger = logging.getLogger(__name__)


class Settings:
    def __init__(self):
        self._config = configparser.ConfigParser()

    def load(self, ini_file: Path):
        logger.debug(f'Loading settings from {ini_file}')
        self._config.read(ini_file)

    @property
    def hotreload_self(self) -> bool:
        return self._config.getboolean('general', 'hotreload_self', fallback=False)

    @property
    def open_url_code(self) -> str:
        return self._config.get('general', 'open_url_code', fallback='')

    @property
    def log_level(self) -> dict[str, str]:
        if not self._config.has_section('log_level'):
            return {}
        return dict(self._config.items('log_level'))



from __future__ import annotations

import types
from types import SimpleNamespace

from wwwpy.common.rpc2.encoder_decoder import EncoderDecoder
from wwwpy.common.rpc2.stub import Stub
from wwwpy.common.rpc2.transport import Transport
from wwwpy.common.rpc2.typed_function import TypedFunction, get_typed_function
from wwwpy.exceptions import RemoteException, RemoteError


class DefaultStub(Stub):

    def __init__(self, transport: Transport, encdec: EncoderDecoder, module_name: str):
        self.namespace = SimpleNamespace()
        self._module_name = module_name
        self._transport = transport
        self._encdec = encdec

    def setup_functions(self, *functions: types.FunctionType) -> None:
        for f in functions:
            self._add_function(f)

    def setup_classes(self, *classes: type) -> None:
        ...

    def _add_function(self, f):
        ft = get_typed_function(f)

        def fun_sync(*args):
            return self.invoke_sync(ft, args)

        async def fun_async(*args):
            return await self.invoke_async(ft, args)

        fun = fun_async if ft.is_coroutine else fun_sync
        setattr(self.namespace, ft.func_name, fun)

    def invoke_sync(self, target_function: TypedFunction, args) -> any:
        send_buffer = self._encode_request(target_function, args)

        self._transport.send_sync(send_buffer)

        recv_buffer = self._transport.recv_sync()
        decode = self._decode_result(target_function, recv_buffer)
        return decode

    async def invoke_async(self, target_function: TypedFunction, args) -> any:
        send_buffer = self._encode_request(target_function, args)

        await self._transport.send_async(send_buffer)
        recv_buffer = await self._transport.recv_async()

        decode = self._decode_result(target_function, recv_buffer)
        return decode

    def _decode_result(self, target_function, recv_buffer):
        decoder = self._encdec.decoder(recv_buffer)
        status = decoder.decode(str)
        if status == 'ex':
            exception = decoder.decode(str)
            raise RemoteException(exception)
        elif status == 'ok':
            decode = decoder.decode(target_function.return_type)
            return decode
        else:
            raise RemoteError(f'Unknown status: `{status}`')

    def _encode_request(self, target_function, args):
        encoder = self._encdec.encoder()
        encoder.encode(self._module_name, str)
        encoder.encode(target_function.func_name, str)
        for arg, arg_type in zip(args, target_function.args_types):
            encoder.encode(arg, arg_type)
        send_buffer = encoder.buffer
        return send_buffer



from remote import rpc

from wwwpy.server.convention import default_project


async def send_message_to_all(msg: str) -> str:
    for client in default_project().websocket_pool.clients:
        client.rpc(rpc.Rpc).new_message(msg)
    return 'done'



import os
import logging
from pathlib import Path


async def main(rootpath, invocation_dir, args):
    from wwwpy.common.designer import log_emit
    def emit(msg: str):
        print(msg)
        from js import console
        console.log(msg)

    log_emit.add_once(emit)
    log_emit.warning_to_log()
    logging.getLogger().setLevel(logging.DEBUG)
    # from wwwpy.remote.designer import log_redirect
    # log_redirect.redirect_logging()

    from js import console
    console.log(f'main({rootpath}, {invocation_dir}, {args})')
    from wwwpy.common.tree import print_tree
    print_tree('/wwwpy_bundle')

    Path('/wwwpy_bundle/pytest.ini').write_text("[pytest]\n"
                                                "asyncio_mode = auto")
    from wwwpy.remote import micropip_install
    await micropip_install('pytest==7.2.2')  # didn't work with update to 8.1.1
    await micropip_install('pytest-asyncio')
    await micropip_install('pytest-xvirt')
    import wwwpy.common.designer as des
    for package in des.pypi_packages:
        await micropip_install(package)
    import pytest
    print('-=-' * 20 + 'pytest imported')

    os.chdir(invocation_dir)
    pytest.main(args)



from wwwpy.common.rpc.serializer import RpcRequest
from wwwpy.websocket import SendEndpoint


class Proxy:
    def __init__(self, module_name: str, send_endpoint: SendEndpoint):
        self.send_endpoint = send_endpoint
        self.module_name = module_name

    def dispatch(self, func_name: str, *args) -> None:
        rpc_request = RpcRequest.to_json(self.module_name, func_name, *args)
        self.send_endpoint.send(rpc_request)



from __future__ import annotations

import asyncio
import configparser
import importlib.metadata
import sys
from pathlib import Path
from typing import Tuple

from wwwpy.common.designer.packaging.packages import Package, PackageSpecification
from wwwpy.common.result import Result


class PackageManager:

    async def installed_packages(self) -> Result[list[Package], Exception]:
        try:
            return Result.success(self._installed_packages_sync())
        except Exception as e:
            return Result.failure(e)

    def _installed_packages_sync(self) -> list[Package]:
        packages = []
        unique = set()
        for dist in importlib.metadata.distributions():
            package = Package(name=dist.metadata['Name'], version=dist.version)
            if package not in unique:
                unique.add(package)
                packages.append(package)

        return packages

    async def _install_package(self, specification: PackageSpecification) -> Exception | None:
        raise NotImplementedError()

    async def install_package_list(self, specifications: list[PackageSpecification | str]) -> Tuple[
        list[Package], Exception | None]:
        """Install all specified packages, stop on the first error"""
        packages = []
        for spec in specifications:
            result = await self.install_package(spec)
            if result.is_failure:
                return packages, result.error_or_null()
            packages.append(Package(name=spec.name, version=''))
        return packages, None

    async def install_package(self, specification: PackageSpecification | str) -> Exception | None:
        spec = PackageSpecification(specification) if isinstance(specification, str) else specification
        return await self._install_package(spec)

    # def update_package(self, specification: PackageSpecification) -> Result[None, Exception]: ...

    async def uninstall_package(self, package: Package | str) -> Exception | None:
        raise NotImplementedError()

    # def reload_packages(self) -> Result[list[Package], Exception]: ...


class PipPackageManager(PackageManager):
    """Pip package manager, using pip as a module, not with subprocess"""

    async def _install_package(self, specification: PackageSpecification) -> Exception | None:
        save_argv = sys.argv
        save_exit = sys.exit
        exit_args = []

        def _sys_exit(arg0):
            exit_args.append(arg0)

        sys.exit = _sys_exit
        try:
            import runpy
            sys.argv = ['pip', 'install'] + specification.build_installation_string()
            runpy.run_module("pip", run_name="__main__")
            if exit_args and exit_args[0] == 0:
                return None
            return Exception(f"pip install failed with exit code {exit_args}")
        except Exception as e:
            return e
        finally:
            sys.argv = save_argv
            sys.exit = save_exit


class UvPackageManager(PackageManager):

    async def _install_package(self, specification: PackageSpecification) -> Exception | None:
        try:
            arg = specification.build_installation_string()
            cmd = ["uv", "pip", "install"] + arg
            proc = await asyncio.create_subprocess_exec(*cmd, stdout=asyncio.subprocess.PIPE,
                                                        stderr=asyncio.subprocess.PIPE)
            stdout, stderr = await proc.communicate()
            if proc.returncode == 0:
                return None
            return Exception(f"uv pip install failed with exit code {proc.returncode}: {stderr.decode()}")
        except Exception as e:
            return e

    async def uninstall_package(self, package: Package | str) -> Exception | None:
        try:
            package_name = package.name if isinstance(package, Package) else package
            cmd = ["uv", "pip", "uninstall", package_name]
            proc = await asyncio.create_subprocess_exec(*cmd, stdout=asyncio.subprocess.PIPE,
                                                        stderr=asyncio.subprocess.PIPE)
            stdout, stderr = await proc.communicate()
            if proc.returncode == 0:
                return None
            return Exception(f"uv pip uninstall failed with exit code {proc.returncode}: {stderr.decode()}")
        except Exception as e:
            return e


class MicropipPackageManager(PackageManager):
    async def _install_package(self, specification: PackageSpecification) -> Exception | None:
        try:
            import micropip
            await micropip.install(specification.build_installation_string())
            return None
        except Exception as e:
            return e

    async def uninstall_package(self, package: Package | str) -> Exception | None:
        try:
            package_name = package.name if isinstance(package, Package) else package
            import micropip
            await micropip.uninstall(package_name)
        except Exception as e:
            return e
def guess_package_manager() -> PackageManager:
    return guess_package_manager_class().get_or_throw()()


def guess_package_manager_class() -> Result[type[PackageManager], ExceptionGroup[...]]:
    exceptions = []
    try:
        import pip
        return Result.success(PipPackageManager)
    except Exception as e:
        exceptions.append(e)

    try:
        import micropip
        return Result.success(MicropipPackageManager)
    except Exception as e:
        exceptions.append(e)

    try:
        _is_uv_managed()
        return Result.success(UvPackageManager)
    except Exception as e:
        exceptions.append(e)

    return Result.failure(ExceptionGroup(*exceptions))


def _is_uv_managed():
    cfg_path = Path(sys.prefix) / 'pyenv.cfg'
    if not cfg_path.is_file():
        return False

    content = "[dummy]\n" + cfg_path.read_text()
    config = configparser.ConfigParser()
    config.read_string(content)

    return 'uv' in config['dummy']



import dataclasses
import shutil
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Dict, Iterable, Optional

from wwwpy.common import tree
from wwwpy.common.filesystem.sync import Event


def events_invert(fs: Path, events: List[Event]) -> List[Event]:
    root = Node('.', '.', True)

    """This is the root node of the tree that will be used to keep track of the path changes"""

    def _get_or_create_node(path_str: str) -> Node:
        path_node = _get_node_chain(root, path_str)[-1]
        if path_node is None:
            is_dir = (fs / path_str).is_dir()
            path_node = _create_node(root, path_str, is_dir)
        return path_node

    def get_final_path(event: Event) -> Path:
        path_node = _get_or_create_node(event.src_path)
        assert path_node is not None
        return Path(fs / path_node.final_path)

    def augment(event: Event) -> Event:
        fp = get_final_path(event)
        content = _get_content(fp)
        aug = dataclasses.replace(event, content=content)
        return aug

    def is_deleted_entity(rel: Event) -> bool:
        chain = _get_node_chain(root, rel.src_path)
        for node in chain:
            if node is not None and node.is_deleted:
                return True
        return False

    def is_modified_entity(rel: Event) -> bool:
        node = _get_node_chain(root, rel.src_path)[-1]
        if node is not None and node.is_modified:
            return True
        return False

    relative_events = []
    for e in reversed(events):
        if e.event_type == 'closed':
            continue
        rel = e.relative_to(fs)
        if rel.src_path == '' or rel.src_path == '.':
            continue  # skip any event on the root
        if is_deleted_entity(rel):
            continue
        # we are processing the events backwards in time to go from A_n to A_0
        if e.event_type == 'moved':
            current_path = rel.dest_path  # this is the path in A_i
            new_path = rel.src_path  # this is the path in A_(i-1)
            final_node = _get_node_chain(root, current_path)[-1]
            if final_node is None:
                # this is the first rename of this path
                _create_node(root, current_path, (fs / current_path).is_dir())

            _move_node(root, current_path, new_path)  # we mutate the tree backwards in time

        if is_modified_entity(rel):
            continue
        if rel.event_type == 'deleted':
            # mark this entity such as we are ignore all preceding events, 'moved' included
            _get_or_create_node(rel.src_path).is_deleted = True
        elif rel.event_type == 'modified':
            if get_final_path(rel).is_dir():
                continue
            rel = augment(rel)
            # mark this entity such as we are ignore all preceding events, except 'moved'
            _get_or_create_node(rel.src_path).is_modified = True

        relative_events.insert(0, rel)

    return relative_events


def events_apply(fs: Path, events: List[Event]):
    for event in events:
        _event_apply(fs, event)


def _event_apply(fs: Path, event: Event):
    path = fs / event.src_path
    t = event.event_type
    is_dir = event.is_directory
    if t == 'created':
        if is_dir:
            path.mkdir(exist_ok=True)  # todo this is because of dev_mode on rpc signature change
        else:
            path.touch()
    elif t == 'deleted':
        if path.exists():  # it could not exist because of events compression
            if is_dir:
                shutil.rmtree(path)  # again because of events compression we could need to remove a whole tree
            else:
                path.unlink()
    elif t == 'moved':
        shutil.move(path, fs / event.dest_path)
    elif t == 'modified':
        c = event.content
        func = None
        if isinstance(c, str):
            func = path.write_text
        elif isinstance(c, bytes):
            func = path.write_bytes
        else:
            raise ValueError(f"Unsupported content type: {type(c)}")

        path.parent.mkdir(parents=True, exist_ok=True)
        func(c)

def _get_content(path: Path):
    assert path.exists(), f'Path does not exist: {path}'
    assert path.is_file(), f'Not a file: {path}'
    try:
        return path.read_text()
    except UnicodeDecodeError:
        return path.read_bytes()


@dataclass
class Node:
    name: str
    """This is the instant name of the node. It is the name of the node in the A_i state"""
    final_path: str
    """This is the final path of the node in the A_n state. It is a relative complete path"""
    is_directory: bool
    is_deleted: bool = False
    is_modified: bool = False
    children: Dict[str, 'Node'] = field(default_factory=dict)

    # validate in post init
    def __post_init__(self):
        if '//' in self.final_path:
            raise ValueError(f'Invalid path: {self.final_path}')

    def str(self):
        return f'{self.name}'


def _get_node_chain(root: Node, path: str) -> List[Optional[Node]]:
    if path == '.':
        return [root]
    parts = _get_parts(path)
    current = root
    chain = [root]

    for part in parts:
        if current is None or part not in current.children:
            chain.append(None)
            current = None
        else:
            current = current.children[part]
            chain.append(current)

    return chain


def _get_parts(path):
    return path.strip("/").split("/")


def _create_node(root: Node, final_path: str, is_dir: bool) -> Node:
    if final_path == '':
        return root
    parts = _get_parts(final_path)
    current = root

    for i, name in enumerate(parts):
        if name not in current.children:
            path = current.final_path + '/' + name
            current.children[name] = Node(name, path, is_dir)
        current = current.children[name]

    return current


def _move_node(root: Node, current_path: str, new_path: str):
    """This function moves a node from the current path to the new path.
    if the new path already exists, we will raise an error"""
    print('\nbefore move')
    print_node(root)
    current_node_chain = _get_node_chain(root, current_path)
    new_node_chain = _get_node_chain(root, new_path)

    if new_node_chain[-1] is not None:
        raise ValueError(f'Path already exists: {new_path}')

    current_node = current_node_chain[-1]
    current_node_parent = current_node_chain[-2]
    new_node_parent = new_node_chain[-2]
    new_name = _get_parts(new_path)[-1]

    current_node_parent.children.pop(current_node.name)
    current_node.name = new_name
    new_node_parent.children[new_name] = current_node

    print('after move')
    print_node(root)
    return


@dataclass
class _NodePrint(tree.NodeProtocol):
    node: Node

    def iterdir(self) -> Iterable[tree.NodeProtocol]:
        return [_NodePrint(child) for child in self.node.children.values()]

    def is_dir(self) -> bool:
        return self.node.is_directory

    @property
    def name(self) -> str:
        msg = f' ({self.node.final_path}) is_dir={self.node.is_directory} is_deleted={self.node.is_deleted}'
        return self.node.str().strip('/') + msg


def print_node(node: Node, printer=print):
    for line in tree.tree(_NodePrint(node)):
        printer(line)


def events_init(source: Path) -> List[Event]:
    result = []
    for path in source.rglob('*'):
        src_path = str(path.relative_to(source))
        if path.is_dir():
            result.append(Event(event_type='created', src_path=src_path, is_directory=True))
        else:
            result.append(Event(event_type='modified', src_path=src_path, is_directory=False,
                                content=_get_content(path)))
    return result



from __future__ import annotations

from typing import Any

from wwwpy.common.rpc.serializer import RpcRequest, RpcResponse
from wwwpy.common.rpc.v2.dispatcher import Dispatcher
from wwwpy.exceptions import RemoteException


# todo cleanup, this is 'mostly' used by the remote for the remote-to-server rpc(s)
#  but it's also used by the server tests. This should take care mostly about serialization/deserialization
#  and error handling, so we may completely abstract the transport (urllib, browser fetch etc etc)
#  and remove old tests and do new one for the new version
#  There are two counterparts to the dispatch, it's RpcRoute.dispatch() and setup_websocket().message()
class HybridDispatcher(Dispatcher):
    def __init__(self, module_name: str, rpc_url: str):
        self.rpc_url = rpc_url
        from wwwpy.common.fetch import async_fetch_str
        self.fetch = async_fetch_str
        self.module_name = module_name

    async def dispatch_async(self, func_name: str, *args) -> Any:
        rpc_request = RpcRequest.to_json(self.module_name, func_name, *args)
        json_response = await self.fetch(self.rpc_url, method='POST', data=rpc_request)
        response = RpcResponse.from_json(json_response)
        ex = response.exception
        if ex is not None and ex != '':
            raise RemoteException(ex)
        return response.result

    def dispatch_sync(self, func_name: str, *args) -> Any:
        rpc_request = RpcRequest.to_json(self.module_name, func_name, *args)
        import js
        xhr = js.XMLHttpRequest.new()
        xhr.open('POST', self.rpc_url, False)
        xhr.setRequestHeader('Content-Type', 'application/json')
        xhr.send(rpc_request)
        json_response = xhr.responseText
        response = RpcResponse.from_json(json_response)
        ex = response.exception
        if ex is not None and ex != '':
            raise RemoteException(ex)
        return response.result



import errno
import logging
import socket

logger = logging.getLogger(__name__)

_start = (10035 - 1)  # windows doesn't like the previous port range
_end = 30000


def find_port() -> int:
    global _start
    while _start < _end:
        _start += 1
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                s.bind(('0.0.0.0', _start))
                print(f'findport() -> {_start}')
                return _start
            except socket.error as e:
                if e.errno != errno.EADDRINUSE:
                    raise

    raise Exception(f'find_port(start={_start},end={_end}) no bindable tcp port found in interval')

if __name__ == '__main__':
    print(find_port())
    print(find_port())


def is_port_busy(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            if hasattr(socket, "SO_REUSEPORT"):
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.bind(('0.0.0.0', port))
            return False
        except socket.error as e:
            # logger.error(f'is_port_busy({port}) -> {e.errno}', exc_info=e)
            if e.errno == errno.EADDRINUSE:
                return True
            raise



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Callable, List, Optional
from contextlib import contextmanager


class Monitorable:
    def __init__(self):
        self.monitor_object = Monitor()


# todo rename to AttributeChanged
@dataclass
class PropertyChanged:
    instance: any
    name: str
    old_value: object
    new_value: object
    origin: Optional[any] = None


@dataclass
class _OriginEvent:
    value: any
    origin: any


# in other situations we could make this a class to be inherited from, so to handle the
# general mechanism of monitoring changes
@dataclass
class Monitor:
    listeners: list[Callable[[List[PropertyChanged]], None]] = field(default_factory=list)
    attr_listeners: dict[str, List[Callable[[List[PropertyChanged]], None]]] = field(default_factory=dict)
    grouping: Optional[List[PropertyChanged]] = None
    origin: Optional[any] = None

    def add_attribute_listener(self, attr_name: str, listener: Callable[[List[PropertyChanged]], None]):
        if attr_name not in self.attr_listeners:
            self.attr_listeners[attr_name] = []
        self.attr_listeners[attr_name].append(listener)

    def notify(self, changes: List[PropertyChanged]):

        if self.grouping is not None:
            self.grouping.extend(changes)
            return

        for l in self.listeners:
            l(changes)
        d = _group_by_attr_name(changes)
        for attr_name, listeners in self.attr_listeners.items():
            if attr_name in d:
                for l in listeners:
                    l(d[attr_name])


def _group_by_attr_name(changes: List[PropertyChanged]):
    d = defaultdict(list)
    for u in changes:
        d[str(u.name)].append(u)
    return dict(d)


__instance_monitor_attr = "__instance_monitor_attr"


def has_monitor(instance):
    return get_monitor(instance) is not None


def get_monitor(instance) -> Optional[Monitor]:
    if isinstance(instance, Monitorable):
        return instance.monitor_object
    return getattr(instance, __instance_monitor_attr, None)


def get_monitor_or_create(instance) -> Monitor:
    m = get_monitor(instance)
    if m is not None:
        return m
    clazz = instance.__class__

    if not hasattr(clazz, "__attr_change_monitor"):
        setattr(clazz, "__attr_change_monitor", True)

        original_setattr = clazz.__setattr__  # Keep a reference to the original method

        def new_setattr(self, name, value):
            old_value = getattr(self, name, None)
            original_setattr(self, name, value)

            monitor = get_monitor(self)
            if monitor is None:
                return
            change = PropertyChanged(self, name, old_value, value, monitor.origin)
            monitor.notify([change])

        clazz.__setattr__ = new_setattr

    m = Monitor()
    instance.__instance_monitor_attr = m
    return m


def get_monitor_or_raise(instance) -> Monitor:
    m = get_monitor(instance)
    if m is None:
        raise Exception("No monitor found")
    return m


# todo should return un unsubscribe function
def monitor_changes(instance, on_changed: Callable[[List[PropertyChanged]], None]):
    """Monitor the changes of the properties of an instance of a class."""

    m = get_monitor_or_create(instance)
    m.listeners.append(on_changed)


@contextmanager
def group_changes(instance):
    # what happens with nested groupings?

    m: Monitor = get_monitor_or_raise(instance)
    buffer = []
    m.grouping = buffer

    try:
        yield instance
    finally:
        m.grouping = None
        if buffer:
            m.notify(buffer)


@contextmanager
def set_origin(instance, origin: any):
    monitor = get_monitor_or_raise(instance)
    monitor.origin = origin
    try:
        yield instance
    finally:
        monitor.origin = None



from __future__ import annotations

import ast
import sys
from dataclasses import dataclass, field
from typing import List, Dict


@dataclass
class Attribute:
    name: str
    type: str
    default: str


@dataclass
class Method:
    name: str
    def_lineno: int
    code_lineno: int
    is_async: bool = field(default=False)


@dataclass
class ClassInfo:
    name: str
    attributes: List[Attribute]
    methods: List[Method] = field(default_factory=list)

    methods_by_name: Dict[str, Method] = field(init=False)
    attributes_by_name: Dict[str, Attribute] = field(init=False)

    def __post_init__(self):
        self.methods_by_name = {method.name: method for method in self.methods}
        self.attributes_by_name = {attr.name: attr for attr in self.attributes}

    def next_attribute_name(self, base_name):
        fixed = _kebab_to_camel(base_name)
        used = set([attr.name for attr in self.attributes])
        for i in range(1, sys.maxsize):
            name = f'{fixed}{i}'
            if name not in used:
                return name
        raise ValueError('Really?')

def _kebab_to_camel(s: str) -> str:
    parts = s.split('-')
    return parts[0] + ''.join(word.capitalize() for word in parts[1:])


def class_info(python_source: str, class_name: str) -> ClassInfo | None:
    classes = info(python_source).classes
    filtered_classes = [clazz for clazz in classes if clazz.name == class_name]
    if len(filtered_classes) == 0:
        return None
    return filtered_classes[0]


@dataclass
class Info:
    classes: List[ClassInfo]


def info(python_source: str) -> Info:
    tree = ast.parse(python_source)
    extractor = _ClassInfoExtractor()
    extractor.visit(tree)
    return Info(extractor.classes)


class _ClassInfoExtractor(ast.NodeVisitor):
    def __init__(self):
        self.classes = []

    def visit_ClassDef(self, node):
        class_name = node.name
        attributes = []
        methods = []
        for item in node.body:
            if isinstance(item, ast.AnnAssign):
                attr_name = item.target.id
                attr_type = self.get_annotation_type(item.annotation)
                default = ast.unparse(item.value) if item.value else None
                attributes.append(Attribute(attr_name, attr_type, default))
            elif isinstance(item, ast.FunctionDef):
                methods.append(Method(item.name, item.lineno, item.body[0].lineno))
            elif isinstance(item, ast.AsyncFunctionDef):
                methods.append(Method(item.name, item.lineno, item.body[0].lineno, True))
        self.classes.append(ClassInfo(class_name, attributes, methods))
        self.generic_visit(node)

    def get_annotation_type(self, annotation):
        if isinstance(annotation, ast.Name):
            return annotation.id
        elif isinstance(annotation, ast.Attribute):
            names = []
            while isinstance(annotation, ast.Attribute):
                names.append(annotation.attr)
                annotation = annotation.value
            if isinstance(annotation, ast.Name):
                names.append(annotation.id)
            names.reverse()
            return '.'.join(names)
        else:
            return 'Unknown'



import asyncio
import logging

logger = logging.getLogger(__name__)


async def rpctst_echo(msg: str) -> str:
    return f'echo {msg}'


async def rpctst_exec(source: str):
    """Example:
    page.mouse.click(100, 100)
    page.locator('input').fill('foo1')

    This method queue the command to be executed in the playwright thread,
    it waits the completion but do not return the result.
    """
    from wwwpy.server.pytestlib.xvirt_impl import xvirt_instances
    from wwwpy.server.pytestlib.playwrightlib import PlaywrightBunch
    i = len(xvirt_instances)
    assert i == 1, f'len(xvirt_instances)={i}'
    xvirt = xvirt_instances[0]
    args = xvirt.playwright_args
    assert args is not None
    pwb: PlaywrightBunch = args.instance
    assert pwb is not None
    gl = {'page': pwb.page}
    loop = asyncio.get_running_loop()
    done = asyncio.Event()
    args.queue.put(lambda: [exec(source, gl), loop.call_soon_threadsafe(done.set)])
    await done.wait()



from __future__ import annotations

from wwwpy.common.asynclib import OptionalCoroutine
from wwwpy.common.rpc.serializer import RpcRequest
from wwwpy.http import HttpRoute, HttpRequest, HttpResponse
from wwwpy.webserver import Route
from wwwpy.websocket import WebsocketRoute, WebsocketEndpoint, ListenerProtocol


# Route = Union[HttpRoute, WebsocketRoute]

def routes_to_asgi_application(*routes: Route):
    return AsgiApplication(routes)


class AsgiApplication:
    def __init__(self, *routes: Route):
        def _g(type_):
            return _groupby_to_dict(filter(lambda r: isinstance(r, type_), routes), lambda i: i.path)

        self.http_route: dict[str, HttpRoute] = _g(HttpRoute)
        self.websocket_route: dict[str, WebsocketRoute] = _g(WebsocketRoute)

        self._scopes = {
            'http': self._scope_http,
            'websocket': self._scope_websocket,
            'lifespan': self._scope_lifespan,
        }

    async def __call__(self, scope, receive, send):
        func = self._scopes.get(scope['type'], None)
        if func is None:
            return
        await func(scope, receive, send)  # noqa

    async def _scope_lifespan(self, scope, receive, send):
        pass

    async def _scope_http(self, scope, receive, send):
        route = self.http_route.get(scope['path'], None)
        if route is None:
            return
        method = scope['method']
        headers = dict(scope['headers'])
        content_type = headers.get('content-type', None)
        body = await _all_body(receive)
        # todo (?) intercept content type to correctly transform body bytes to str if needed
        http_request = HttpRequest(method, body, content_type)

        def resp_callback(resp: HttpResponse) -> OptionalCoroutine:
            async def future():
                await send({'type': 'http.response.start', 'status': 200,
                            'headers': [[b'content-type', resp.content_type.encode()], ], })
                await send({'type': 'http.response.body', 'body': resp.content.encode(), })

            return future()

        res = route.callback(http_request, resp_callback)
        if res:
            await res

    async def _scope_websocket(self, scope, receive, send):
        route = self.websocket_route.get(scope['path'], None)
        if route is None:
            return

        # send_text = lambda t: send({'type': 'websocket.send', 'text': t})
        async def send_text(t):
            await send({'type': 'websocket.send', 'text': t})

        await send({'type': 'websocket.accept'})

        endpoint = _AsgiWebsocketEndpoint(lambda m: send_text(m))
        route.on_connect(endpoint)

        while True:
            message = await receive()
            if message['type'] == 'websocket.receive':
                await endpoint.on_message(message.get('text'))
            elif message['type'] == 'websocket.disconnect':
                break


def _groupby_to_dict(iterable, key) -> dict:
    result = dict()
    for item in iterable:
        k = key(item)
        array = result.get(k, None)
        if array is None:
            array = []
            result[k] = array
        array.append(item)
    return result


async def _all_body(receive):
    body = b""
    more_body = True

    while more_body:
        message = await receive()
        body += message.get("body", b"")
        more_body = message.get("more_body", False)
    return body


class _AsgiWebsocketEndpoint(WebsocketEndpoint):
    def __init__(self, send: ListenerProtocol):
        self._send = send
        self._listener: ListenerProtocol | None = None

    def add_listener(self, listener: ListenerProtocol) -> None:
        if self._listener is not None:
            raise Exception('Only one listener is allowed')
        self._listener = listener

    # part to be called by user code to send a outgoing message
    def send(self, message: str | bytes | None) -> OptionalCoroutine:
        return self._send(message)

    # parte to be used by IO implementation to be called to notify incoming messages
    def on_message(self, message: str | bytes | None) -> OptionalCoroutine:
        if self._listener is not None:
            return self._listener(message)

    def dispatch(self, module: str, func_name: str, *args) -> OptionalCoroutine:
        j = RpcRequest.to_json(module, func_name, *args)
        return self.send(j)



from __future__ import annotations

import logging

from wwwpy.common.rpc2.transport import Transport

logger = logging.getLogger(__name__)

try:
    import js


    class RemoteHttpTransport(Transport):
        def __init__(self, rpc_url: str):
            self.rpc_url = rpc_url
            self.response = None

        def _buf(self, payload: str | bytes):
            if self.response:
                raise Exception('Cannot send twice with this implementation')
            self.response = payload

        def _consume(self):
            if self.response is None:
                raise Exception('Cannot consume before sending')
            response = self.response
            self.response = None
            return response

        async def send_async(self, payload: str | bytes):
            logger.debug(f'send_async payload: {payload}')
            json_response = await js.fetch(self.rpc_url, method='POST', body=payload)
            text = await json_response.text()
            # response = RpcResponse.from_json(json_response)
            # ex = response.exception
            # if ex is not None and ex != '':
            #     raise RemoteException(ex)
            self._buf(text)

        def send_sync(self, payload: bytes):
            # rpc_request = RpcRequest.to_json(self.module_name, func_name, *args)
            # import js
            xhr = js.XMLHttpRequest.new()
            xhr.open('POST', self.rpc_url, False)
            xhr.setRequestHeader('Content-Type', 'application/json')
            xhr.send(payload)
            json_response = xhr.responseText
            # response = RpcResponse.from_json(json_response)
            # ex = response.exception
            # if ex is not None and ex != '':
            #     raise RemoteException(ex)
            self._buf(json_response)

        def recv_sync(self) -> bytes:
            return self._consume()

        async def recv_async(self) -> str | bytes:
            return self._consume()


    class ServerHttpTransport(Transport):
        pass
except:
    class ServerHttpTransport(Transport):
        def __init__(self, request: str | bytes):
            self._recv = request
            self.response = None

        def _buf(self, payload: str | bytes):
            if self.response:
                raise Exception('Cannot send twice with this implementation')
            self.response = payload

        def _consume_recv(self):
            if self._recv is None:
                raise Exception('Cannot consume before sending')
            response = self._recv
            self._recv = None
            return response

        def recv_sync(self) -> bytes:
            return self._consume_recv()

        async def recv_async(self) -> str | bytes:
            return self._consume_recv()

        def send_sync(self, payload: bytes):
            self._buf(payload)

        async def send_async(self, payload: str | bytes):
            self._buf(payload)


    class RemoteHttpTransport(Transport):
        ...



import logging
from dataclasses import dataclass

import js
import wwwpy.remote.component as wpc
from wwwpy.common.state import _restore

from .mailto_component import MailtoComponent

logger = logging.getLogger(__name__)


@dataclass
class State:
    recipient: str = ''
    subject: str = ''
    body: str = ''
    text_content: str = ''


class MailtoEditComponent(wpc.Component):
    _recipient: js.HTMLElement = wpc.element()
    _subject: js.HTMLElement = wpc.element()
    _body: js.HTMLElement = wpc.element()
    _text_content: js.HTMLElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<h1>MailtoEditComponent</h1>
<sl-input data-name="_recipient" placeholder="Recipient" label="Recipient" ></sl-input>
<sl-input data-name="_subject" placeholder="Subject" label="Subject"></sl-input>
<sl-textarea data-name="_body" placeholder="Body" label="Body">slTextarea1</sl-textarea>
<sl-input data-name="_text_content" placeholder="Link text content" label="Link text content"></sl-input>
<br>
        """
        self.state = _restore(State)
        self._recipient.value = self.state.recipient
        self._subject.value = self.state.subject
        self._body.value = self.state.body
        self._text_content.value = self.state.text_content

        target = MailtoComponent()
        self.element.append(target.element)
        self.target = target
        self._update()

    def _update(self):
        self.state.recipient = self._recipient.value
        self.state.subject = self._subject.value
        self.state.body = self._body.value
        self.state.text_content = self._text_content.value

        self.target.recipient = self._recipient.value
        self.target.body = self._body.value
        self.target.subject = 'add_component error report'
        self.target.text_content = self._text_content.value

    async def _recipient__sl_input(self, event):
        self._update()

    async def _subject__sl_input(self, event):
        self._update()

    async def _body__sl_input(self, event):
        self._update()

    async def _text_content__sl_input(self, event):
        self._update()

    async def _text_content__click(self, event):
        js.console.log('handler _text_content__click event =', event)



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Union

import js
from pyodide.ffi import create_proxy

import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js
from wwwpy.remote.designer.global_interceptor import InterceptorEvent, GlobalInterceptor


@dataclass
class Actions:
    set_input_value = True
    hide_dropdown = True
    dispatch_change_event = True


class Option:
    parent: SearchableComboBox

    def __init__(self, text: str = ''):
        self.text = text
        self.actions = Actions()
        self.on_selected = lambda: None

    def _on_click(self, event: js.MouseEvent):
        js.console.log('_on_click', self.text)
        if self.actions.set_input_value:
            self.set_input_value()
        if self.actions.hide_dropdown:
            self.hide_dropdown()
        if self.actions.dispatch_change_event:
            self.dispatch_change_event()

        self.on_selected()

    def set_input_value(self):
        self.parent.input.value = self.text

    def hide_dropdown(self):
        self.parent.hide_dropdown()

    def dispatch_change_event(self):
        self.parent._dispatch_change_event()

    def create_element(self) -> js.HTMLElement:
        element = self.new_element()
        element.addEventListener('click', create_proxy(self._on_click))
        return element

    def new_element(self) -> js.HTMLElement:
        div: js.HTMLDivElement = js.document.createElement('div')
        div.textContent = self.text
        return div


class SearchableComboBox(wpc.Component, tag_name='wwwpy-searchable-combobox'):
    input: js.HTMLInputElement = wpc.element()
    search: js.HTMLInputElement = wpc.element()
    dropdown: js.HTMLElement = wpc.element()
    popup: js.HTMLElement = wpc.element()

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
        <style>
    :host {
        display: inline-block;
        position: relative;
        font-family: Arial, sans-serif;
    }
    input {
        width: 90%;
        padding: 5px;
        background-color: #2a2a2a;
        color: #e0e0e0;
        border: 1px solid #444;
        border-radius: 4px;
        font-size: 14px;
    }
    input::placeholder {
        color: #888;
    }
    .dropdown {
        position: absolute;
        width: 100%;
        max-height: 200px;
        resize: both;
        overflow-y: auto;
        border: 1px solid #444;
        background-color: #333;
        color: #e0e0e0;
        border-radius: 0 0 4px 4px;
        box-shadow: 0 4px 6px rgba(0, 0, 0, 0.3);
    }
    .popup {
        position: absolute;
        z-index: 1000;
        display: none;
    }   
    .dropdown * {
        padding: 8px 12px;
        cursor: pointer;
        transition: background-color 0.2s ease;
    }
    .dropdown div:hover {
        background-color: #444;
    }
    input:focus {
        outline: none;
        border-color: #666;
        box-shadow: 0 0 0 2px rgba(100, 100, 100, 0.3);
    }
</style>
        <input type="text" placeholder="" data-name="input">
        <div data-name="popup" class="popup">
            <input type="search" placeholder="Search..." data-name="search">
            <div class="dropdown" data-name="dropdown"></div>
        </div>
        """
        self._interceptor = GlobalInterceptor(self._global_click)
        self._options = []
        self._dirty = False
        self.hide_dropdown()

    def _global_click(self, event: InterceptorEvent):
        click_inside = self.element.contains(event.target)
        js.console.log(f'global click: contains={click_inside}', event.target)
        if click_inside:
            return
        self.hide_dropdown()

    def search__input(self, event):
        self.filter_options()

    def input__change(self, event):
        if self._is_free_edit():
            self._dispatch_change_event()

    def input__mousedown(self, event: js.MouseEvent):
        if self._is_free_edit():
            return
        event.preventDefault()

    @property
    def _search_visible(self):
        return self.popup.style.display != 'none'

    def input__click(self, event: js.MouseEvent):
        if self._search_visible:
            self.hide_dropdown()
        else:
            self.show_dropdown()

    def show_dropdown(self):
        if self._is_free_edit():
            return
        if self._dirty:
            self.filter_options()
        self.popup.style.display = 'block'
        self.search.focus()
        self._interceptor.install()

    def _is_free_edit(self):
        return not self._options

    def hide_dropdown(self):
        self.popup.style.display = 'none'
        self._interceptor.uninstall()

    def set_options(self, options: List[Union[str, Option]]):
        self._options = options
        self._dirty = True

    def filter_options(self, event=None):
        self._dirty = False
        filter_text = self.search.value.lower()
        self.dropdown.innerHTML = ''

        for option in self._options:
            if isinstance(option, str):
                option = Option(option)
            option.parent = self
            option: Option
            if filter_text in option.text.lower():
                self.dropdown.appendChild(option.create_element())

        if not self.dropdown.hasChildNodes():
            div: js.HTMLDivElement = js.document.createElement('div')
            div.textContent = 'No results'
            div.style.textAlign = 'center'
            div.style.fontStyle = 'italic'
            div.style.fontWeight = 'bold'
            div.style.pointerEvents = 'none'
            self.dropdown.appendChild(div)

    def select_option(self, option):
        js.console.log(f'option: {option}')
        self.input.value = option
        self.hide_dropdown()
        self._dispatch_change_event()

    def _dispatch_change_event(self):
        self.element.dispatchEvent(js.CustomEvent.new('wp-change', {'detail': self.input.value}))



from __future__ import annotations

import importlib
import inspect
from dataclasses import dataclass, field
from typing import Any

from wwwpy.common.designer.html_locator import NodePath, locate_span
from wwwpy.common.modlib import _find_module_root
from wwwpy.common import modlib
from enum import Enum


class Origin(str, Enum):
    source = 'source'
    """The original html source in the component"""
    live = 'live'
    """The live html in the browser"""


@dataclass()
class ElementPath:
    """Contains the path to an element relative to a Component.
    This is intended to be serialized"""

    class_module: str
    """The module name of the Component."""
    class_name: str
    """The class name of the Component."""
    path: NodePath
    """The path from the Component (excluded) to the element."""

    origin: Origin

    @property
    def tag_name(self) -> str:
        """The tag name of the element in lowercase."""
        if len(self.path) == 0:
            return ''
        return self.path[-1].tag_name

    @property
    def data_name(self) -> Optional[str]:
        """The name of the element in the data dictionary."""
        if len(self.path) == 0:
            return None
        return self.path[-1].attributes.get('data-name', None)

    def valid(self) -> bool:
        from wwwpy.common.designer import code_strings as cs, html_parser as hp, html_locator as hl
        html = cs.html_from(self.class_module, self.class_name)
        if not html:
            return False
        return hl.locate_span(html, self.path) is not None



import threading
from datetime import timedelta
from pathlib import Path
from time import sleep
from typing import Callable, List

from watchdog.events import FileSystemEvent

from wwwpy.common.filesystem.sync import new_tmp_path
from wwwpy.common.filesystem.sync.event import Event
from wwwpy.server.filesystem_sync.any_observer import AnyObserver
from wwwpy.server.filesystem_sync.debouncer import Debouncer
from wwwpy.server.filesystem_sync.debouncer_thread import DebouncerThread


class WatchdogDebouncer(DebouncerThread):

    def __init__(self, path: Path, window: timedelta, callback: Callable[[List[Event]], None]):
        self._debouncer = Debouncer(window)
        super().__init__(self._debouncer, callback)

        def skip_open(event: FileSystemEvent):
            if event.event_type == 'opened' or event.event_type == 'closed_no_write':
                return
            if event.is_synthetic:
                return

            e = Event(event.event_type, event.is_directory, event.src_path, event.dest_path)
            self._debouncer.add_event(e)

        self._any_observer = AnyObserver(path, skip_open)

    @property
    def debouncer(self):
        return self._debouncer

    def start(self):
        self._any_observer.watch_directory()
        super().start()

    def stop(self):
        self._any_observer.stop()
        super().stop()

    def join(self):
        self._any_observer.join()
        super().join()

    def watch_directory(self):
        self.start()


def main():
    tmp_path = new_tmp_path()
    print(f'watching {tmp_path}')

    def print_events(events: List[Event]):
        print('=' * 20 + f' current thread name={threading.current_thread().name}')
        print(f'originals:')
        for e in events:
            print(f'  {e}')
        print('relative:')
        for e in events:
            e = e.relative_to(tmp_path)
            print(f'  {e}')
        print(f'len={len(events)}')

    WatchdogDebouncer(tmp_path, timedelta(milliseconds=100), print_events).start()

    while True:
        sleep(1)


if __name__ == '__main__':
    main()



from __future__ import annotations

from typing import List, Union, Callable

import wwwpy.remote.component as wpc
import js
from wwwpy.remote import dict_to_js
from pyodide.ffi import create_proxy
import logging

logger = logging.getLogger(__name__)

class ButtonTab(wpc.Component, tag_name='wwwpy-button-tab'):
    _root: js.HTMLElement = wpc.element()

    @property
    def tabs(self) -> List[Tab]:
        return self._tabs

    @tabs.setter
    def tabs(self, value: List[Union[str, Tab]]):
        def process(v):
            if not isinstance(v, Tab):
                v = Tab(v)
            v.parent = self
            return v

        self._tabs = [process(v) for v in value]
        re = self._root
        re.innerHTML = ''
        for tab in self._tabs:
            re.appendChild(tab.root_element())

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
<style>
  .selected {
        box-shadow: 0 0 0 2px darkgray;
        }
</style>
<span data-name="_root"></span>
"""
        style = self._root.style
        style.display = 'flex'
        style.justifyContent = 'space-around'
        style.width = '100%'

        self._tabs = []
        self._updating = False

    def _tab_selected(self, tab: Tab):
        if self._updating:
            return
        self._updating = True
        for t in self._tabs:
            t.selected = t == tab
        self._updating = False

_selected_default = lambda tab: None


class Tab:
    parent: ButtonTab

    def __init__(self, text: str = '', on_selected: Callable[[Tab], None] = _selected_default):
        self.text = text
        self.on_selected = on_selected
        self._root_element: js.HTMLElement = None

    def do_click(self, *event):
        self.selected = True
        self.on_selected(self)

    def root_element(self) -> js.HTMLElement:
        if self._root_element is None:
            div = js.document.createElement('button')
            div.textContent = self.text
            div.addEventListener('click', create_proxy(self.do_click))
            self._root_element = div
        return self._root_element

    @property
    def selected(self) -> bool:
        return self.root_element().classList.contains('selected')

    @selected.setter
    def selected(self, value: bool):
        current = self.selected
        class_list = self.root_element().classList
        if value:
            class_list.add('selected')
        else:
            class_list.remove('selected')

        self.parent._tab_selected(self)



from dataclasses import dataclass
from typing import Protocol


@dataclass(frozen=True)
class FunctionDef:
    name: str
    annotations: list[type]
    return_annotation: type


@dataclass(frozen=True)
class Definition:
    target: str
    functions: dict[str, FunctionDef]


class Dispatcher(Protocol):
    """
    This Caller/Callee convention this is the Caller.
    https://en.wikipedia.org/wiki/Calling_convention

    This is used by the proxy_generator to be used inside the generated source code.
    The proxy_generator will instantiate a Dispatcher for the module and one for each class.

    It requires the __module__ and __qualname__ attributes. because they will be they will be imported
    in the generated source code.

    For example, classes and functions have these attributes.
    """
    __module__: str
    __qualname__: str

    def definition_complete(self, definition: Definition) -> None:
        """The proxy_generator will call this method when the top level module
    is parsed and also at the end of each class definition. This allows the implementation to
    inspect the functions and their type hints through the locals() dictionary."""
        ...

    def dispatch_sync(self, function_name: str, *args) -> any:
        """The proxy_generator will call this method to dispatch a function call to the implementation."""
        ...

    async def dispatch_async(self, function_name: str, *args) -> any:
        """The proxy_generator will call this method to dispatch a function call to the implementation."""
        ...



from typing import Callable, Awaitable

from wwwpy.common.asynclib import OptionalCoroutine
from wwwpy.http import HttpRoute
from wwwpy.server.asgi import routes_to_asgi_application, AsgiApplication
from wwwpy.webserver import Webserver, Route
from wwwpy.websocket import WebsocketRoute


class AsgiWebserver(Webserver):
    def __init__(self, start):
        super().__init__()
        self._start = start
        self.app = AsgiApplication()

    def _setup_route(self, route: Route):
        if isinstance(route, HttpRoute):
            self.app.http_route[route.path] = route
        elif isinstance(route, WebsocketRoute):
            self.app.websocket_route[route.path] = route
        else:
            raise Exception(f'Unknown route type: {type(route)}')

    async def _start_listen(self) -> None:
        await self._start(self)



from __future__ import annotations

import ast
import types
from typing import Optional

from wwwpy.common.rpc2.transport import Transport


class JsonStub:
    def __init__(self, dispatcher: Transport):
        self.dispatcher = dispatcher


class Stub:
    """
    See naming conventions at https://en.wikipedia.org/wiki/Distributed_object_communication

    This is class is placed inside the generated source code, so it MUST be available on the Caller side.
    It will be instantiated when the module is loaded on the Caller side and not at generation time.

    It requires the __module__ and __qualname__ attributes. because they will be used to generate the import statement
    and the instantiation in the generated source code; classes and functions have these attributes.
    """
    __module__: str
    __qualname__: str

    namespace: any
    """This is the namespace that will be used to forward the function/method calls to the Stub implementation    
    The namespace needs to take care of creating the correct sync or async function/method calls
    """

    def setup_functions(self, *functions: types.FunctionType) -> None:
        """This must setup the namespace for the functions"""

    def setup_classes(self, *classes: type) -> None:
        """This must setup the namespace for the classes"""


_annotations_type = set[ast.Name]

_stub_name = '_stub'

def generate_stub(source: str, stub_type: type[Stub], stub_args: str = '') -> str:
    """Generates a stub source code from the given source code.
This implementation is divides the stub in two parts, one the proxy code generated by this function and the other
part is a concrete implementation of Stub that is called by the proxy code.

The goals are:
- forward the function/method calls to the Stub instance
- preserve the type hints and default parameters of the functions/methods
- preserve the import and importFrom statements so the type hints are available
- removes the implementation of the functions and replaces it with a forwarding call to the stub

Inclusion/Exclusion
- MUST NOT generate entities (function/method/class) that starts with '_'
    """

    tree: ast.Module = ast.parse(source)
    module = stub_type.__module__
    qualified_name = stub_type.__qualname__
    lines = [
        f'from {module} import {qualified_name}',
        f'{_stub_name} = {qualified_name}({stub_args})',
        ''
    ]
    used_annotations: _annotations_type = set()
    function_names = []
    class_names = []
    for b in tree.body:
        if isinstance(b, (ast.FunctionDef, ast.AsyncFunctionDef)) and not b.name.startswith('_'):
            _add_function_or_method(lines, b, used_annotations)
            function_names.append(b.name)
        elif isinstance(b, (ast.ImportFrom, ast.Import)):
            lines.append(b)
        elif isinstance(b, ast.ClassDef) and not b.name.startswith('_'):
            class_names.append(b.name)
            class_body = b.body
            b.body = []
            class_def = ast.unparse(b)
            lines.append(class_def)
            indent = '    '
            method_count = 0
            for m in class_body:
                if isinstance(m, (ast.FunctionDef, ast.AsyncFunctionDef)) and not m.name.startswith('_'):
                    method_lines = []
                    _add_function_or_method(method_lines, m, used_annotations, class_name=b.name)
                    for ml in method_lines:
                        if ml:
                            ml = indent + ml
                        lines.append(ml)
                    method_count += 1
            if method_count == 0:
                lines.append(indent + 'pass')

    for idx in range(len(lines)):
        line = lines[idx]
        if isinstance(line, ast.Import):
            lines[idx] = ast.unparse(line) if _is_import_used(line, used_annotations) else ''
        elif isinstance(line, ast.ImportFrom):
            lines[idx] = ast.unparse(line) if _is_import_from_used(line, used_annotations) else ''

    # setup_functions call
    lines.append(f'{_stub_name}.setup_functions({", ".join(function_names)})')
    # setup_classes call
    lines.append(f'{_stub_name}.setup_classes({", ".join(class_names)})')

    body = '\n'.join(lines)
    return body


def _add_function_or_method(lines, b, used_annotations, class_name=''):
    b.body = []  # keep only the signature
    func_def = ast.unparse(b)
    lines.append(func_def)
    args_list = []
    anno_list = []
    for ar in b.args.args:
        args_list.append(f'{ar.arg}')
        if ar.annotation:  # this is because of method definition: self is not annotated
            used_annotations.add(ar.annotation)
            anno_list.append(ast.unparse(ar.annotation))
    args = ', '.join(args_list)
    if b.returns:
        used_annotations.add(b.returns)
    is_async = isinstance(b, ast.AsyncFunctionDef)
    async_spec = 'await ' if is_async else ''
    name = b.name
    if class_name:
        name = f'{class_name}.{name}'
    lines.append(f'    return {async_spec}{_stub_name}.namespace.{name}({args})')
    lines.append('')  # empty line after each function


def _is_import_used(node: ast.Import, used_annotations: _annotations_type) -> bool:
    for alias in node.names:
        candidate = alias.asname if alias.asname is not None else alias.name
        for ann in used_annotations:
            if _annotation_uses_candidate(ann, candidate):
                return True
    return False


def _is_import_from_used(node: ast.ImportFrom, used_annotations: _annotations_type) -> bool:
    for alias in node.names:
        candidate = alias.asname if alias.asname is not None else alias.name
        for ann in used_annotations:
            if _annotation_uses_candidate(ann, candidate):
                return True
    return False


def _annotation_uses_candidate(node: ast.AST, candidate: str) -> bool:
    full_name = _get_full_name(node)
    if full_name is not None:
        if full_name == candidate or full_name.startswith(candidate + '.'):
            return True
    for child in ast.iter_child_nodes(node):
        if _annotation_uses_candidate(child, candidate):
            return True
    return False


def _get_full_name(node: ast.AST) -> Optional[str]:
    if isinstance(node, ast.Name):
        return node.id
    elif isinstance(node, ast.Attribute):
        base = _get_full_name(node.value)
        if base is not None:
            return base + '.' + node.attr
    return None


# language=python
"""
from some_module import SomeThing # noqa
_stub = Stub(__name__, stub_args) # noqa

def add(a: int, b: int) -> int:
    return _stub.namespace.add(a, b)

async def sub(a: int, b: int) -> SomeThing:
    return await _stub.namespace.sub(a, b)

class Class1:
    
    def __init__(self):
        return _stub.namespace.Class1.__init__(self)
    
    def add(self, c: int) -> int:
        return _stub.namespace.Class1.add(self, c)

class Class2:
        
    async def sub(self, c: int) -> int:
        return await _stub.namespace.Class1.sub(self, c)
    
            
_stub.setup_functions(add, sub)
_stub.setup_classes(Class1, Class2)
"""



from __future__ import annotations

import asyncio
import logging
from typing import Callable, Optional, Dict, Union, Awaitable

import js
from js import console, window, EventTarget, KeyboardEvent
from pyodide.ffi import create_proxy, to_js

logger = logging.getLogger(__name__)

HotkeyHandler = Union[Callable[['KeyboardEvent'], Optional[bool]], Callable[['KeyboardEvent'], Awaitable[None]]]


class Hotkey:
    def __init__(self, element: EventTarget):
        self.element = element
        self.handlers: Dict[str, HotkeyHandler] = dict()
        self.debug = False
        self._proxy = create_proxy(self._detect_hotkey)
        self.install()

    def install(self):
        self.element.addEventListener('keydown', self._proxy, False)

    def uninstall(self):
        self.element.removeEventListener('keydown', self._proxy, False)

    # todo this should be called 'set'
    def add(self, hotkey: str, handler: HotkeyHandler) -> 'Hotkey':
        """
        Registers a hotkey with its corresponding handler.

        Adds a key-handler pair to the handlers dictionary, associating the hotkey with
        the given handler function or callable.

        Parameters
        ----------
        hotkey : str
            The key combination that will trigger the handler. Examples: 'CTRL-S', 'Escape', 'META-Backspace'
            , 'CTRL-SHIFT-ALT-META-F1', this last example serves to know the order of the modifiers.
        handler : HotkeyHandler
            A function or callable that will be called when the hotkey is pressed.
        """
        self.handlers[hotkey] = handler
        return self

    @classmethod
    def keyboard_event(cls, e):
        return js.eval('(e) => e instanceof KeyboardEvent ')(e)

    def _detect_hotkey(self, e):
        if not self.keyboard_event(e):
            return
        key = ''
        if e.ctrlKey: key += 'CTRL-'
        if e.shiftKey: key += 'SHIFT-'
        if e.altKey: key += 'ALT-'
        if e.metaKey: key += 'META-'

        upc: str = e.key
        if len(upc) == 1: upc = upc.upper()

        key += upc
        if self.debug:
            console.log(key, to_js(e))
            logger.debug(f'key = `{key}`')

        handle = self.handlers.get(key, None)
        if handle is None:
            return
        e.hotkey = key  # add the hotkey attribute to the event
        if asyncio.iscoroutinefunction(handle):
            e.preventDefault()
            e.stopPropagation()
            asyncio.create_task(handle(e))
            return

        res = handle(e)
        if not res:
            return

        console.log(f'prevent default for {key}')
        e.preventDefault()
        e.stopPropagation()



from __future__ import annotations

import asyncio
import logging

import js

logger = logging.getLogger(__name__)


async def activate():
    from wwwpy.remote import micropip_install
    from wwwpy.common import designer

    for package in designer.pypi_packages:
        await micropip_install(package)

    from wwwpy.remote.designer import helpers
    js.window.onerror = helpers._on_error
    js.window.onunhandledrejection = helpers._on_unhandledrejection
    from wwwpy.remote.designer import log_redirect
    log_redirect.redirect_logging()

    loop = asyncio.get_event_loop()
    loop.set_exception_handler(_global_exception_handler)


def is_active() -> bool:
    try:
        return js.window._wwwpy_dev_mode
    except AttributeError:
        return False


def _global_exception_handler(loop, context):
    # The context parameter contains details about the exception
    if 'message' in context:
        logger.info(f"Global handler caught: {context['message']}")
    if 'exception' not in context:
        logger.error(f"Global handler caught: {context}")
        return
    exception = context.get('exception')
    if exception:
        logger.info(f"Exception type: {type(exception)}, Args: {exception.args}")
        logger.exception(exception)


def set_active(dev_mode: bool):
    js.window._wwwpy_dev_mode = dev_mode



from __future__ import annotations

import asyncio
import logging
from typing import Dict

import js
from js import HTMLElement, console
from pyodide.ffi import create_proxy

logger = logging.getLogger(__name__)
namespace = "window.python_custom_elements"


# todo maybe we can join Metadata and Component into one class
class Metadata:
    def __init__(self, tag_name: str | None = None, clazz=None):
        if clazz is not None:
            if not issubclass(clazz, Component):
                raise Exception(f'clazz must be a subclass of {Component.__name__}')
            if tag_name is None:
                # get the full class name
                fully_qualified_class_name = clazz.__qualname__
                tag_name = ('wwwpy-auto-' + fully_qualified_class_name.lower()
                            .replace('<', '-')
                            .replace('>', '-')
                            .replace('.', '-')
                            )

        self.tag_name = tag_name
        self.observed_attributes = set()
        self.registered = False
        self.clazz = clazz
        self._custom_element_class_template = None

    def __set_name__(self, owner, name):
        if not issubclass(owner, Component):
            raise Exception(f'attribute {name} must be in a subclass of {Component.__name__}')
        self.clazz = owner
        # if self.auto_define:
        #     self.define_element()

    def define_element(self):
        if self.registered:
            return
        if js.eval(f'typeof {namespace}') == 'undefined':
            js.eval(namespace + ' = {}')

        js_class_name = self.tag_name.replace('-', '_').replace('.', '_')
        self._js_class_name = js_class_name

        pc = js.eval(namespace)
        already_defined = hasattr(pc, js_class_name)
        setattr(pc, js_class_name, self.clazz)  # set the python constructor, in any case
        if not already_defined:
            obs_attr = ', '.join(f'"{attr}"' for attr in self.observed_attributes)
            code = (_custom_element_class_template
                    .replace('$ClassName', js_class_name)
                    .replace('$tagName', self.tag_name)
                    .replace('$observedAttributes', obs_attr)
                    .replace('$namespace', namespace)
                    )
            self._custom_element_class_template = code
            # raise Exception(code)
            js.eval(code)

        self.registered = True

    @property
    def html_snippet(self):
        """Returns the complete HTML markup for this custom element."""
        return self.build_snippet()

    def build_snippet(self, attributes: Dict[str, str | None] = None, inner_html: str = '') -> str:
        """Returns a snippet of HTML for this custom element."""
        if attributes is None:
            attributes = {}
        attr_str = ' '.join(f'{k}="{v}"' for k, v in attributes.items())
        return f'<{self.tag_name} {attr_str}>{inner_html}</{self.tag_name}>'


def get_component(element: HTMLElement) -> Component | None:
    if not hasattr(element, '_python_component'):
        return None
    component = element._python_component
    if hasattr(component, "unwrap"):
        component = component.unwrap()

    return component


# PUBLIC-API
class Component:
    component_metadata: Metadata = None
    element: HTMLElement = None
    element_not_found_raises = False

    def __init_subclass__(cls, tag_name: str | None = None, auto_define=True, **kwargs):
        super().__init_subclass__(**kwargs)
        metadata = Metadata(tag_name=tag_name, clazz=cls)
        cls.component_metadata = metadata
        if metadata.clazz is None:
            metadata.clazz = cls

        for name, value in cls.__dict__.items():
            if isinstance(value, attribute):
                cls.component_metadata.observed_attributes.add(value.name)

        if auto_define:
            cls.component_metadata.define_element()

    def __init__(self, element_from_js=None):
        if element_from_js is None:
            self.element = js.eval(f'window.{self.component_metadata._js_class_name}').new(create_proxy(self))
        else:
            self.element = element_from_js

        self.init_component()
        self._bind_events()

        if asyncio.iscoroutinefunction(self.after_init_component):
            asyncio.create_task(self.after_init_component())
        else:
            self.after_init_component()

    async def after_init_component(self):
        """This is called after init_component, it can be async or called synchronously if it is a normal method.
        It is called after the event binding is completed"""
        pass

    def init_component(self):
        pass

    def connectedCallback(self):
        pass

    def disconnectedCallback(self):
        pass

    def adoptedCallback(self):
        pass

    def attributeChangedCallback(self, name: str, oldValue: str, newValue: str):
        pass

    def root_element(self):
        """This is used to locate the child elements"""
        e = self.element
        return e if e.shadowRoot is None else e.shadowRoot

    def _find_element(self, name: str):
        root = self.root_element()
        selector = root.querySelector(f'[data-name="{name}"]')
        if selector is None:
            msg = f'Not found data-name: [{name}] html: [{root.outerHTML}]'
            if self.element_not_found_raises:
                raise ElementNotFound(msg)
            else:
                console.warn(msg)

        return selector

    def _find_python_attribute(self, name: str):
        selector = self._find_element(name)

        if selector:
            comp = get_component(selector)
            if comp:
                selector = comp

        self._check_type(name, selector)
        return selector

    def _check_type(self, name, selector):
        import inspect
        try:
            annotations = inspect.get_annotations(self.__class__, eval_str=True)
        except Exception as e:
            fqn = f'{self.__module__}.{self.__class__.__name__}'
            import sys
            state_unloaded = self.__module__ not in sys.modules
            att_ds = (f'Class {fqn} was unloaded but an instance is still referenced (with e.g., through DOM events)'
                      f'\nAttribute accessed `{name}` Instance id={id(self)}'
                      f'\nTo fix this error, remove the listeners that are firing (usually in a `disconnectedCallback` method)'
                      f'\nBeware that the stack trace could point to the wrong line because the instance in memory'
                      f'\nwas generated by a previous source.'
                      f'\nStack trace:\n{_get_stack_trace_string(-4)}') if state_unloaded else ''
            if not att_ds:
                logger.error(f'Error {fqn} getting annotations name=`{name}` selector=`{selector}`: {e}')
            else:
                try:
                    from wwwpy.server.designer import rpc
                    asyncio.create_task(rpc.on_exception_string(att_ds))
                except:
                    logger.error(att_ds)

            if logger.isEnabledFor(logging.DEBUG):
                logger.exception(e)
            return

        expected_type = annotations.get(name, None)
        if expected_type is None:
            # raise Exception(f'No type defined for field: {name}')
            return
        # raise Exception(f'type of expected_type: {type(expected_type)}')
        # test if expected_type is a class
        if not inspect.isclass(expected_type):
            return
        if not issubclass(expected_type, Component):
            return
        # raise Exception(f'Expected type: {expected_type} for field: {name} but found: {type(selector)}')
        isinst = not isinstance(selector, expected_type)
        # raise Exception(f'isinst: {isinst}')
        if isinst:
            raise WrongTypeDefinition(f'Expected type: {expected_type} for field: {name} but found: {type(selector)}')

    def _bind_events(self):

        members = dir(self)

        for name in members:
            if name.startswith('__'):
                continue
            parts = name.split('__')
            if len(parts) != 2:
                continue

            element_name = parts[0]
            event_name = parts[1].replace('_', '-')
            if element_name == '' or event_name == '':
                continue

            element = self._find_element(element_name)
            if element is None:
                console.warn(f'Event bind failed, element `{element_name}` was not found for method `{name}`')
                continue

            m = getattr(self, name)
            element.addEventListener(event_name, create_proxy(m))


class ElementNotFound(AttributeError): pass


class WrongTypeDefinition(TypeError): pass


# language=javascript
_custom_element_class_template = """
class $ClassName extends HTMLElement {
    static observedAttributes = [ $observedAttributes ];
    constructor(python_component) {
        super();
        if (python_component) 
            this._python_component = python_component;
        else 
            this._python_component = $namespace.$ClassName(this);
    }

    connectedCallback()    { this._python_component.connectedCallback(); }
    disconnectedCallback() { this._python_component.disconnectedCallback(); }
    adoptedCallback()      { this._python_component.adoptedCallback(); }
    attributeChangedCallback(name, oldValue, newValue) { this._python_component.attributeChangedCallback(name, oldValue, newValue); }
}

customElements.define('$tagName', $ClassName);
window.$ClassName = $ClassName;
"""


# class Attribute(str):
#     _attr: attribute
#     present: bool
#
#     def toggle(self):
#         pass

# PUBLIC-API
class attribute:

    def __init__(self, name: str | None = None):
        self.name = name

    def __set_name__(self, owner, name):
        if not issubclass(owner, Component):
            raise Exception(f'attribute {name} must be in a subclass of {Component.__qualname__}')
        if self.name is None:
            self.name = name

    def __get__(self, obj: Component, objtype=None):
        return obj.element.getAttribute(self.name)

    def __set__(self, obj: Component, value):
        obj.element.setAttribute(self.name, value)


# PUBLIC-API
class element:

    def __init__(self, cached=False):
        """ If cached is True, the element will be cached after the first access.
        Default is False.
        The practical use of this is when removing the element from the parent, the element is still available.
        Without cache, if it is detached from the parent, the element will be unreachable.
        """
        self.name = None
        self.cached = cached
        self.cache = None

    def __set_name__(self, owner, name):
        if not issubclass(owner, Component):
            raise Exception(f'attribute {name} must be in a subclass of {Component.__qualname__}')
        self.name = name

    def __get__(self, obj: Component, objtype=None):
        value = obj._find_python_attribute(self.name) if self.cache is None else self.cache
        if self.cached and self.cache is None:
            self.cache = value
        return value


def _get_stack_trace_string(skip_frames):
    import traceback
    stack = traceback.extract_stack()[:skip_frames]
    formatted_stack = traceback.format_list(stack)
    return ''.join(formatted_stack)



from typing import Any
import asyncio
from tornado.websocket import WebSocketHandler
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

UTF8 = "utf-8"


class ASGIHandler(WebSocketHandler):

    def initialize(self, asgi_app) -> None:
        super().initialize()
        self._asgi_app = asgi_app

    async def get(self, *args: Any, **kwargs: Any):
        if self.request.headers.get("Upgrade", "").lower() != "websocket":
            await self._handle_http_request(args, kwargs)
            return
        await super().get(*args, **kwargs)  # continue as a real WebSocketHandler, thus to self.open()

    def open(self):
        logger.debug("WebSocket connection opened")
        self.receive_queue = asyncio.Queue()
        asyncio.create_task(self._run_asgi_app())

    async def _run_asgi_app(self):
        scope = {
            "type": "websocket",
            "asgi": {"version": "3.0"},
            "http_version": self.request.version,
            "scheme": self.request.protocol,
            "path": self.request.path,
            "raw_path": self.request.path.encode(UTF8),
            "query_string": self.request.query.encode(UTF8),
            "headers": self._collect_headers(),
            "client": (self.request.remote_ip, 0),
            "server": self.request.host,
            "subprotocols": self.request.headers.get_list("Sec-WebSocket-Protocol"),
        }

        async def receive():
            return await self.receive_queue.get()

        async def send(message):
            message_type = message.get('type')
            if message_type == 'websocket.send':
                if 'text' in message:
                    await self.write_message(message['text'])
                elif 'bytes' in message:
                    await self.write_message(message['bytes'], binary=True)
            elif message_type == 'websocket.close':
                code = message.get('code', 1000)
                reason = message.get('reason', '')
                await self.close(code=code, reason=reason)
            else:
                logger.error(f"Unsupported ASGI message type: {message_type}")

        try:
            await self._asgi_app(scope, receive, send)
        except Exception as e:
            logger.error(f"Error in ASGI application: {e}")
            await self.close()

    def on_message(self, message):
        asyncio.create_task(self._queue_message(message))

    async def _queue_message(self, message):
        await self.receive_queue.put({'type': 'websocket.receive',
                                      'bytes' if isinstance(message, bytes) else 'text': message})

    def on_close(self):
        logger.debug("WebSocket connection closed")
        asyncio.create_task(self._queue_disconnect())

    async def _queue_disconnect(self):
        await self.receive_queue.put({'type': 'websocket.disconnect'})

    async def post(self, *args: Any, **kwargs: Any):
        await self._handle_http_request(args, kwargs)

    async def _handle_http_request(self, *args: Any, **kwargs: Any):
        scope = {
            "type": self.request.protocol,
            "http_version": self.request.version,
            "path": self.request.path,
            "method": self.request.method,
            "query_string": self.request.query.encode(UTF8),
            "headers": self._collect_headers(),
            "client": (self.request.remote_ip, 0),
            "server": self.request.host,
        }

        async def receive():
            return {'body': self.request.body, "type": "http.request", "more_body": False}

        async def send(data):
            if data['type'] == 'http.response.start':
                self.set_status(data['status'])
                self.clear_header("content-type")
                self.clear_header("server")
                self.clear_header("date")
                for h in data['headers']:
                    if len(h) == 2:
                        self.add_header(h[0].decode(UTF8), h[1].decode(UTF8))
            elif data['type'] == 'http.response.body':
                self.write(data['body'])
            else:
                raise RuntimeError(
                    f"Unsupported response type \"{data['type']}\" for ASGI app")

        await self._asgi_app(scope, receive, send)

    def _collect_headers(self):
        headers = []
        for k in self.request.headers:
            for v in self.request.headers.get_list(k):
                headers.append((k.encode(UTF8).lower(), v.encode(UTF8)))
        return headers



from __future__ import annotations


class Transport:

    def send_sync(self, payload: str | bytes): raise NotImplementedError

    async def send_async(self, payload: str | bytes): raise NotImplementedError

    def recv_sync(self) -> str | bytes: raise NotImplementedError

    async def recv_async(self) -> str | bytes: raise NotImplementedError



from __future__ import annotations

import asyncio
from threading import Thread
from typing import Awaitable, Union
from typing import Optional

import tornado
import tornado.web
from tornado import websocket
from tornado.ioloop import IOLoop

from wwwpy.http import HttpRoute, HttpRequest
from ..webserver import Webserver, Route
from ..websocket import WebsocketRoute, WebsocketEndpointIO


class WsTornado(Webserver):
    ioloop: IOLoop = None

    def __init__(self):
        super().__init__()
        self.app = tornado.web.Application()
        self.thread: Optional[Thread] = None

    def _setup_route(self, route: Route):
        if isinstance(route, WebsocketRoute):
            self.app.add_handlers(r".*", [(route.path, _WebsocketHandler, dict(route=route, server=self))])
        else:
            self.app.add_handlers(r".*", [(route.path, TornadoHandler, dict(route=route))])

    def _start_listen(self):
        def run():
            self.app.listen(self.port, self.host)
            self.ioloop = IOLoop.current()
            # asyncio.set_event_loop(self.ioloop.asyncio_loop)
            self.ioloop.start()

        self.thread = Thread(target=run, daemon=True)
        self.thread.start()


class TornadoHandler(tornado.web.RequestHandler):

    def __init__(self, *args, **kwargs):
        self.route: Route = None
        self._serve = None
        super().__init__(*args, **kwargs)

    def initialize(self, route: HttpRoute) -> None:
        self.route = route
        if not isinstance(route, HttpRoute):
            raise Exception(f'Unknown route type: {type(route)}')

    def set_default_headers(self):
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Headers', '*')
        self.set_header('Access-Control-Allow-Methods', '*')

    def options(self):
        """/OPTIONS handler for preflight CORS checks."""
        self.set_status(204)
        self.finish()

    async def get(self) -> None:
        await self._serve_std('GET')

    async def post(self) -> None:
        await self._serve_std('POST')

    async def _serve_std(self, verb: str):
        body = self.request.body
        request = HttpRequest(verb, body, self.request.headers.get('Content-Type', ''))

        def response_fun(response):
            self.set_header("Content-Type", response.content_type)
            return self.write(response.content)

        res = self.route.callback(request, response_fun)
        if res:
            await res

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        raise_exception(self)


class _WebsocketHandler(websocket.WebSocketHandler):
    route: WebsocketRoute = None
    server: WsTornado = None
    endpoint: WebsocketEndpointIO = None

    def initialize(self, route: WebsocketRoute, server: WsTornado) -> None:
        self.route = route
        self.server = server

    # def check_origin(self, origin):
    #     return True

    def open(self):
        self.endpoint = WebsocketEndpointIO(self._on_send)
        self.route.on_connect(self.endpoint)

    def on_message(self, message: Union[str, bytes]) -> Optional[Awaitable[None]]:
        self.endpoint.on_message(message)

    def _on_send(self, data: str | bytes | None):
        if data is None:
            self.close()
        else:
            self.server.ioloop.asyncio_loop.call_soon_threadsafe(self.write_message, data)

    def on_close(self):
        self.endpoint.on_message(None)


def raise_exception(self: TornadoHandler):
    print("=" * 60)
    print(f'exc! {self.route}')
    raise Exception(self)



from js import fetch, console

# logger = logging.getLogger(__name__)
logger = console  # if log redirect is active, it will create an infinite loop because

async def async_fetch_str(url: str, method: str = 'GET', data: str = '') -> str:
    logger.debug(f'url={url}')
    logger.debug(f'method={method}')
    logger.debug(f'len(data)={len(data)}')
    logger.debug(f'data=`{data[:100]}`')
    response = await fetch(url, method=method, body=data)
    text = await response.text()
    return text



import asyncio
import logging
import random

import js
from js import document
from pyodide.ffi import create_proxy

logger = logging.getLogger(__name__)


# logger.setLevel('DEBUG')


async def script_load_once(src: str, script_type='', **kwargs) -> bool:
    """This function loads a (javascript) script once. It waits for the script to load and returns True if the script was loaded,
    and False if the script was already loaded."""

    marker = f'm-{random.randint(1000, 9999)}'
    logger.debug(f'{marker} script_load_once: {src} {script_type} {kwargs}')

    scripts = list(filter(lambda s: s.src == src, document.head.querySelectorAll('script')))  # noqa
    scripts: list[js.HTMLScriptElement]
    need_load = not scripts
    script = None if need_load else scripts[0]

    if script:
        logger.debug(f'{marker}   found script with same src: {src} type={script.type}')
        assert script.type == script_type, f'script.type=`{script.type}`, script_type=`{script_type}`'
    else:
        logger.debug(f'{marker}  load is needed, no script with same src found: {src}')
        script = document.createElement('script')
        # copy all kwargs to the script
        for key, value in kwargs.items():
            setattr(script, key, value)
        script.src = src
        if script_type:
            script.type = script_type
        script.__async_event = asyncio.Event()
        script.onload = create_proxy(lambda e: script.__async_event.set())
        document.head.append(script)

    logger.debug(f'{marker}  waiting for script to load: {src}')
    await script.__async_event.wait()
    logger.debug(f'{marker}  event set: need_load={need_load} {src}')
    return need_load
